# -*- coding: utf-8 -*-
""" Implement the evaluating requests section of the
GraphQL specification. """

# REVIEW: The resolver interface is mirrored on graphql-js but that might change

# REVIEW: The Future based interface can be awkward especially when checking
# for Future instances in order to gain performance.

import functools as ft

import six

from . import _concurrency
from .._string_utils import stringify_path
from .._utils import OrderedDict, find_one, is_iterable, lazy
from ..exc import (
    CoercionError,
    ExecutionError,
    ResolverError,
    ScalarSerializationError,
    UnknownEnumValue,
)
from ..lang import ast as _ast
from ..schema import (
    EnumType,
    IncludeDirective,
    InterfaceType,
    ListType,
    NonNullType,
    ObjectType,
    ScalarType,
    Schema,
    SkipDirective,
    UnionType,
    is_abstract_type,
)
from ..schema.introspection import schema_field, type_field, type_name_field
from ..utilities import (
    coerce_argument_values,
    coerce_variable_values,
    default_resolver,
    directive_arguments,
)
from .executors import DefaultExecutor, Executor
from .middleware import apply_middlewares
from .wrappers import ExecutionContext, GraphQLResult, ResolveInfo


def execute(  # flake8: noqa : C901
    schema,
    ast,
    executor=None,
    variables=None,
    operation_name=None,
    initial_value=None,
    context_value=None,
    middlewares=None,
):
    """ Execute a graphql query against a schema.

    Args:
        schema (py_gql.schema.Schema): Schema to execute the query against

        ast (py_gql.lang.ast.Document): The parsed query

        executor (Optional[py_gql.execution.Executor]): Executor instance

        variables (Optional[dict]):
            Raw, JSON decoded variables parsed from the request

        operation_name (Optional[str]): Operation to execute
            If specified, the operation with the given name will be executed.
            If not, this executes the single operation without disambiguation.

        initial_value (Any): Root resolution value passed to all top-level
            resolvers

        context_value (Any): Custom application-specific execution context
            Use this to pass in anything your resolvers require like database
            connection, user information, etc.
            Limits on the type(s) used here will depend on your own resolver
            implementations and the executor class you use. Most thread safe
            data-structures should work.

        middlewares (Optional[List[Callable]]):
            List of middleware callable to use when resolving fields
    """
    # Programmer errors
    assert isinstance(schema, Schema), "Expected Schema object"
    assert isinstance(ast, _ast.Document), "Expected document"
    assert variables is None or isinstance(
        variables, dict
    ), "Variables must be a dictionnary"
    assert executor is None or isinstance(executor, Executor)
    assert not middlewares or all(callable(mw) for mw in middlewares)

    executor = executor or DefaultExecutor()
    variables = variables or dict()

    operation = get_operation(ast, operation_name)
    object_type = {
        "query": schema.query_type,
        "mutation": schema.mutation_type,
        "subscription": schema.subscription_type,
    }[operation.operation]

    if object_type is None:
        raise ExecutionError(
            "Schema doesn't support %s operation" % operation.operation
        )

    fragments = {
        d.name.value: d
        for d in ast.definitions
        if d.__class__ is _ast.FragmentDefinition
    }

    coerced_variables = coerce_variable_values(schema, operation, variables)

    ctx = ExecutionContext(
        schema,
        ast,
        coerced_variables,
        fragments,
        executor,
        operation,
        middlewares,
        context_value,
    )

    if operation.operation == "query":
        deferred_result = _execute_selections(
            ctx, operation.selection_set.selections, object_type, initial_value
        )
    elif operation.operation == "mutation":
        deferred_result = _execute_selections_serially(
            ctx, operation.selection_set.selections, object_type, initial_value
        )
    else:
        raise NotImplementedError("%s not supported" % operation.operation)

    def _on_end(data):
        return GraphQLResult(
            data=data, errors=[err for err, _, _ in ctx.errors]
        )

    if _concurrency.is_deferred(deferred_result):
        return _concurrency.chain(deferred_result, _on_end, cls=ctx.future_cls)
    return _concurrency.deferred(_on_end(deferred_result), cls=ctx.future_cls)


def get_operation(document, operation_name):
    """ Extract relevant operation from a parsed document

    :type document: py_gql.lang.ast.Document
    :param document: Parsed document

    :type operation_name: Optional[str]
    :param operation_name: Operation name to extract

    :rtype: py_gql.lang.ast.OperationDefinition
    """
    operations = [
        definition
        for definition in document.definitions
        if definition.__class__ is _ast.OperationDefinition
    ]

    if not operations:
        raise ExecutionError("Expected at least one operation")

    if not operation_name:
        if len(operations) == 1:
            return operations[0]
        else:
            raise ExecutionError(
                "Operation name is required when document "
                "contains multiple operation definitions"
            )

    operation = find_one(operations, lambda o: o.name.value == operation_name)
    if operation is not None:
        return operation
    else:
        raise ExecutionError(
            'No operation "%s" found in document' % operation_name
        )


def collect_fields(
    ctx, object_type, selections, visited_fragments=None
):  # noqa : C901
    """ Collect all fields in a selection set, recursively traversing fragments
    in one single map and conserving definitino order.

    :type ctx: ExecutionContext
    :param ctx:
        Current execution context

    :type object_type: py_gql.schema.Type
    :param object_type:
        Current object type

    :type selections: List[py_gql.lang.ast.Selection]
    :param selections:
        Selections from the SelectionSet node(s) to gather fields from

    :type visited_fragments: Optional[Set[str]]
    :param visited_fragments:
        List of already visited fragment spreads

    :rtype: OrderedDict
    """

    cache_key = object_type.name, tuple(selections)
    if cache_key in ctx.grouped_fields:
        return ctx.grouped_fields[cache_key]

    visited_fragments = visited_fragments or set()
    grouped_fields = OrderedDict()

    def _collect_fragment_fields(fragment_selections):
        fragment_grouped_fields = collect_fields(
            ctx, object_type, fragment_selections, visited_fragments
        )
        for key, gf in fragment_grouped_fields.items():
            if key not in grouped_fields:
                grouped_fields[key] = []
            grouped_fields[key].extend(gf)

    for selection in selections:
        kind = selection.__class__

        if kind is _ast.Field:
            if not _include_selection(selection, ctx.variables):
                continue

            key = (
                selection.alias.value
                if selection.alias
                else selection.name.value
            )
            if key not in grouped_fields:
                grouped_fields[key] = []
            grouped_fields[key].append(selection)

        elif kind is _ast.InlineFragment:
            if not _include_selection(selection, ctx.variables):
                continue

            if not fragment_type_applies(ctx.schema, selection, object_type):
                continue

            _collect_fragment_fields(selection.selection_set.selections)

        elif kind is _ast.FragmentSpread:
            if not _include_selection(selection, ctx.variables):
                continue

            name = selection.name.value
            if name in visited_fragments:
                continue

            fragment = ctx.fragments.get(name)

            if not fragment_type_applies(ctx.schema, fragment, object_type):
                continue

            _collect_fragment_fields(fragment.selection_set.selections)
            visited_fragments.add(name)

    ctx.grouped_fields[cache_key] = grouped_fields
    return grouped_fields


def fragment_type_applies(schema, fragment, object_type):
    """ Determines if a fragment is applicable to the given type.

    :type schema: py_gql.schema.Schema
    :param schema:
        Current schema

    :type fragment: py_gql.lang.ast.InlineFragment|\
        py_gql.lang.ast.FragmentDefinition
    :param fragment:
        Fragment node

    :type object_type: py_gql.schema.Type
    :param object_type:
        Current object type

    :rtype: bool
    """
    type_condition = fragment.type_condition
    if not type_condition:
        return True

    fragment_type = schema.get_type_from_literal(type_condition)
    if fragment_type == object_type:
        return True

    if is_abstract_type(fragment_type):
        return schema.is_possible_type(fragment_type, object_type)

    return False


def _include_selection(node, variables=None):
    skip = directive_arguments(SkipDirective, node, variables=variables)
    include = directive_arguments(IncludeDirective, node, variables=variables)
    skipped = skip is not None and skip["if"]
    included = include is None or include["if"]
    return (not skipped) and included


def _field_def(ctx, parent_type, name):

    cache_key = parent_type.name, name

    if cache_key not in ctx.field_defs:
        if name in ("__schema", "__type", "__typename"):
            if name == "__schema" and ctx.schema.query_type == parent_type:
                ctx.field_defs[cache_key] = schema_field
            elif name == "__type" and ctx.schema.query_type == parent_type:
                ctx.field_defs[cache_key] = type_field
            elif name == "__typename":
                ctx.field_defs[cache_key] = type_name_field
            else:
                ctx.field_defs[cache_key] = parent_type.field_map.get(
                    name, None
                )
        else:
            ctx.field_defs[cache_key] = parent_type.field_map.get(name, None)

    return ctx.field_defs[cache_key]


def _resolve_type(value, context, schema, abstract_type):
    if callable(getattr(abstract_type, "resolve_type", None)):
        return abstract_type.resolve_type(value, context=context, schema=schema)
    if isinstance(value, dict) and value.get("__typename__"):
        return value["__typename__"]
    elif hasattr(value, "__typename__") and value.__typename__:
        return value.__typename__
    else:
        possible_types = schema.get_possible_types(abstract_type)
        for type_ in possible_types:
            if callable(getattr(type_, "is_type_of", None)):
                if type_.is_type_of(value, context=context, schema=schema):
                    return type_
        return None


def _execute_selections(ctx, selections, object_type, object_value, path=None):
    fields = collect_fields(ctx, object_type, selections)
    deferred_fields = []
    path = path or []

    for key, nodes in fields.items():
        field_def = _field_def(ctx, object_type, nodes[0].name.value)
        if field_def is None:
            continue

        deferred_fields.append(
            resolve_field(
                ctx,
                object_type,
                object_value,
                field_def,
                nodes,
                path + [key],
                lambda x: x,
            )
        )

    return _concurrency.chain(
        _concurrency.all_(deferred_fields, cls=ctx.future_cls),
        OrderedDict,
        cls=ctx.future_cls,
    )


def _execute_selections_serially(
    ctx, selections, object_type, object_value, path=None
):
    fields = collect_fields(ctx, object_type, selections)
    resolved_fields = []
    steps = []
    path = path or []

    for key, nodes in fields.items():
        field_def = _field_def(ctx, object_type, nodes[0].name.value)
        if field_def is None:
            continue

        steps.append(
            ft.partial(
                resolve_field,
                ctx,
                object_type,
                object_value,
                field_def,
                nodes,
                path + [key],
                resolved_fields.append,
            )
        )

    steps.append(lambda: OrderedDict(resolved_fields))
    return _concurrency.serial(steps, cls=ctx.future_cls)


def resolve_field(
    ctx, parent_type, parent_value, field_def, nodes, path, on_success
):
    """ Execute a field resolver in the current executor and expose
    result as a Future. """

    key = path[-1]
    info = ResolveInfo(
        field_def,
        parent_type,
        path,
        ctx.schema,
        ctx.variables,
        ctx.fragments,
        ctx.operation,
        nodes,
        ctx.executor,
    )

    resolver = field_def.resolve or default_resolver

    def _unwrap(value):
        value = lazy(value)
        if _concurrency.is_deferred(value):
            return _concurrency.unwrap(value, cls=ctx.future_cls)
        return value

    def resolve(root, args, context, info):
        resolved = _unwrap(
            ctx.executor.submit(resolver, root, args, context, info)
        )
        if _concurrency.is_deferred(resolved):
            complete = ft.partial(
                complete_value, ctx, field_def.type, nodes, info.path
            )
            return _concurrency.chain(
                resolved, lazy, complete, cls=ctx.future_cls
            )
        return complete_value(ctx, field_def.type, nodes, info.path, resolved)

    resolve = apply_middlewares(resolve, ctx.middlewares)

    def _on_success(field_value):
        return on_success((key, field_value))

    def _on_error(err):
        ctx.add_error(err, nodes[0], path)
        return on_success((key, None))

    try:
        args = _argument_values(ctx, field_def, nodes)
    except CoercionError as err:
        return _on_error(err)

    try:
        resolved_or_deferred = resolve(parent_value, args, ctx.context, info)
    except (CoercionError, ResolverError) as err:
        return _on_error(err)
    else:
        if _concurrency.is_deferred(resolved_or_deferred):
            return _concurrency.except_(
                _concurrency.chain(
                    resolved_or_deferred, _on_success, cls=ctx.future_cls
                ),
                (CoercionError, ResolverError),
                _on_error,
                cls=ctx.future_cls,
            )
        return _on_success(resolved_or_deferred)


def complete_value(ctx, field_type, nodes, path, resolved_value):
    kind = field_type.__class__
    if kind is NonNullType:
        # REVIEW:
        # - Error is different than ref implementation
        # - Shouldn't this be a RuntimeError ? As in the developer should
        # never return a null non nullable field and instead reais
        # explicitely if the query lead to this behaviour ?
        def _handle_null(value):
            if value is None:
                ctx.add_error(
                    'Field "%s" is not nullable' % stringify_path(path),
                    nodes[0],
                    path,
                )
            return value

        return _concurrency.chain(
            complete_value(ctx, field_type.type, nodes, path, resolved_value),
            _handle_null,
            cls=ctx.future_cls,
        )

    if resolved_value is None:
        return None

    if kind is ListType:
        if not is_iterable(resolved_value, False):
            raise RuntimeError(
                'Field "%s" is a list type and resolved value '
                "should be iterable" % stringify_path(path)
            )

        return _concurrency.all_(
            [
                complete_value(ctx, field_type.type, nodes, path + [i], entry)
                for i, entry in enumerate(resolved_value)
            ],
            cls=ctx.future_cls,
        )

    if kind is ScalarType:
        try:
            serialized = field_type.serialize(resolved_value)
        except ScalarSerializationError as err:
            raise RuntimeError(
                'Field "%s" cannot be serialized as "%s": %s'
                % (path, field_type, err)
            )
        else:
            return serialized

    if kind is EnumType:
        try:
            serialized = field_type.get_name(resolved_value)
        except UnknownEnumValue as err:
            raise RuntimeError(
                'Field "%s" cannot be serialized as "%s": %s'
                % (path, field_type, err)
            )
        else:
            return serialized

    if kind in (ObjectType, InterfaceType, UnionType):
        if kind in (InterfaceType, UnionType):
            runtime_type = _resolve_type(
                resolved_value, ctx.context, ctx.schema, field_type
            )

            if isinstance(runtime_type, six.string_types):
                runtime_type = ctx.schema.get_type(runtime_type, None)

            if not isinstance(runtime_type, ObjectType):
                raise RuntimeError(
                    'Abstract type "%s" must resolve to an ObjectType at runtime '
                    'for field "%s". Received "%s".'
                    % (field_type, path, runtime_type)
                )
            # Backup check in case of badly implemented `resolve_type`,
            # `is_type_of`
            if not ctx.schema.is_possible_type(field_type, runtime_type):
                raise RuntimeError(
                    'Runtime ObjectType "%s" is not a possible type for '
                    'field "%s" of type "%s".'
                    % (runtime_type, stringify_path(path), field_type)
                )
        else:
            runtime_type = field_type

        return _execute_selections(
            ctx, list(_subselections(nodes)), runtime_type, resolved_value, path
        )


def _subselections(nodes):
    for field in nodes:
        for selection in field.selection_set.selections:
            yield selection


def _argument_values(ctx, field_def, nodes):
    cache_key = (field_def, nodes[0])
    if cache_key not in ctx.argument_values:
        ctx.argument_values[cache_key] = coerce_argument_values(
            field_def, nodes[0], ctx.variables
        )
    return ctx.argument_values[cache_key]



# -*- coding: utf-8 -*-

import copy
import functools as ft
import hashlib

import pytest
import six

from py_gql import graphql
from py_gql._string_utils import dedent
from py_gql.exc import ScalarParsingError, SDLError
from py_gql.schema import (
    Argument,
    Directive,
    EnumType,
    EnumValue,
    Field,
    Int,
    ListType,
    NonNullType,
    ObjectType,
    ScalarType,
    String,
)
from py_gql.schema.build import SchemaDirective, make_executable_schema
from py_gql.utilities import default_resolver


def wrap_resolver(field_def, func):
    source_resolver = field_def.resolve or default_resolver

    @ft.wraps(source_resolver)
    def wrapped(parent_value, args, context, info):
        value = source_resolver(parent_value, args, context, info)
        if value is None:
            return value
        return func(value)

    field_def = copy.copy(field_def)
    field_def.resolve = wrapped
    return field_def


def test_simple_field_modifier():
    class UppercaseDirective(SchemaDirective):
        def visit_field(self, field_definition):
            return wrap_resolver(field_definition, lambda x: x.upper())

    assert (
        graphql(
            make_executable_schema(
                """
                directive @upper on FIELD_DEFINITION

                type Query {
                    foo: String @upper
                }
                """,
                schema_directives={"upper": UppercaseDirective},
            ),
            "{ foo }",
            {},
            initial_value={"foo": "lowerCase"},
        ).response()
        == {"data": {"foo": "LOWERCASE"}}
    )


def test_directive_on_wrong_location():
    class UppercaseDirective(SchemaDirective):
        def visit_field(self, field_definition):
            return wrap_resolver(field_definition, lambda x: x.upper())

    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            directive @upper on FIELD_DEFINITION

            type Query @upper {
                foo: String
            }
            """,
            schema_directives={"upper": UppercaseDirective},
        )

    assert exc_info.value.to_dict() == {
        "locations": [{"column": 24, "line": 4}],
        "message": 'Directive "@upper" not applicable to "OBJECT"',
    }


def test_ignores_unknown_directive_implementation():
    make_executable_schema(
        """
        directive @upper on FIELD_DEFINITION

        type Query @upper {
            foo: String
        }
        """
    )


def test_field_modifier_using_arguments():
    class PowerDirective(SchemaDirective):

        definition = Directive(
            "power",
            ["FIELD_DEFINITION"],
            [Argument("exponent", Int, default_value=2)],
        )

        # pylint: disable = super-init-not-called
        def __init__(self, args):
            self.exponent = args["exponent"]

        def visit_field(self, field_definition):
            return wrap_resolver(field_definition, lambda x: x ** self.exponent)

    assert (
        graphql(
            make_executable_schema(
                """
                type Query {
                    foo: Int @power
                    bar: Int @power(exponent: 3)
                }
                """,
                schema_directives={"power": PowerDirective},
            ),
            "{ foo, bar }",
            {},
            initial_value={"foo": 2, "bar": 2},
        ).response()
        == {"data": {"foo": 4, "bar": 8}}
    )


def test_object_modifier_and_field_modifier():
    class UppercaseDirective(SchemaDirective):
        def visit_field(self, field_definition):
            return wrap_resolver(field_definition, lambda x: x.upper())

    class UniqueID(SchemaDirective):
        # pylint:disable=super-init-not-called
        def __init__(self, args):
            self.name = args["name"]
            self.source = args["source"]
            assert len(self.source)

        def resolve(self, root, *args):
            m = hashlib.sha256()
            for fieldname in self.source:
                print(fieldname)
                m.update(str(root.get(fieldname, "")).encode("utf8"))
            return m.hexdigest()

        def visit_object(self, object_definition):
            assert self.name not in object_definition.field_map
            assert all(s in object_definition.field_map for s in self.source)
            return ObjectType(
                name=object_definition.name,
                description=object_definition.description,
                fields=(
                    [Field(self.name, String, resolve=self.resolve)]
                    + object_definition.fields
                ),
                interfaces=object_definition.interfaces,
                is_type_of=object_definition.is_type_of,
                nodes=object_definition.nodes,
            )

    schema = make_executable_schema(
        """
        directive @uid (name: String! = "uid", source: [String]!) on OBJECT
        directive @upper on FIELD_DEFINITION

        type Query {
            foo: Foo
        }

        type Foo @uid (source: ["id", "name"]) {
            id: Int
            name: String @upper
        }
        """,
        schema_directives={"upper": UppercaseDirective, "uid": UniqueID},
    )

    assert schema.to_string() == dedent(
        """
        directive @uid(name: String! = "uid", source: [String]!) on OBJECT

        directive @upper on FIELD_DEFINITION

        type Foo {
            uid: String
            id: Int
            name: String
        }

        type Query {
            foo: Foo
        }
        """
    )

    assert graphql(
        schema,
        "{ foo { uid, name, id } }",
        {},
        initial_value={"foo": {"name": "some lower case name", "id": 42}},
    ).response() == {
        "data": {
            "foo": {
                "uid": (
                    "6dc146d52a9962cfb9b29c2414f68cc628e2a0dcce"
                    "7832760ddf39a441726173"
                ),
                "name": "SOME LOWER CASE NAME",
                "id": 42,
            }
        }
    }


def test_missing_definition():
    class NoopDirective(SchemaDirective):
        pass

    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query {
                foo: String @upper
            }
            """,
            schema_directives={"upper": NoopDirective},
        )

    assert exc_info.value.to_dict() == {
        "locations": [{"column": 29, "line": 3}],
        "message": 'Unknown directive "@upper',
    }


def test_multiple_directives_applied_in_order():
    class PowerDirective(SchemaDirective):

        definition = Directive(
            "power",
            ["FIELD_DEFINITION"],
            [Argument("exponent", Int, default_value=2)],
        )

        # pylint:disable=super-init-not-called
        def __init__(self, args):
            self.exponent = args["exponent"]

        def visit_field(self, field_definition):
            return wrap_resolver(field_definition, lambda x: x ** self.exponent)

    class PlusOneDirective(SchemaDirective):
        definition = Directive("plus_one", ["FIELD_DEFINITION"])

        def visit_field(self, field_definition):
            return wrap_resolver(field_definition, lambda x: x + 1)

    assert (
        graphql(
            make_executable_schema(
                """
                type Query {
                    foo: Int @power @plus_one
                    bar: Int @plus_one @power
                }
                """,
                schema_directives={
                    "power": PowerDirective,
                    "plus_one": PlusOneDirective,
                },
            ),
            "{ foo, bar }",
            {},
            initial_value={"foo": 2, "bar": 2},
        ).response()
        == {"data": {"foo": 5, "bar": 9}}
    )


def test_input_values():
    class LimitedLengthScalarType(ScalarType):
        @classmethod
        def wrap(cls, type_, *args, **kwargs):
            if isinstance(type_, (NonNullType, ListType)):
                return type(type_)(cls.wrap(type_.type, *args, **kwargs))
            return cls(type_, *args, **kwargs)

        # pylint:disable=super-init-not-called,redefined-builtin
        def __init__(self, type, min, max):
            assert isinstance(type, ScalarType)
            self.type = type
            self.min = min
            self.max = max if max is not None else float("inf")
            assert self.min >= 0
            assert self.max >= 0
            self.name = "LimitedLenth%s_%s_%s" % (type.name, self.min, self.max)

        def serialize(self, value):
            return self.type.serialize(value)

        def parse(self, value):
            parsed = self.type.parse(value)
            if not isinstance(parsed, six.string_types):
                raise ScalarParsingError("Not a string")
            if not (self.min <= len(parsed) <= self.max):
                raise ScalarParsingError(
                    "%s length must be between %s and %s but was %s"
                    % (self.name, self.min, self.max, len(parsed))
                )
            return parsed

    class LimitedLengthDirective(SchemaDirective):

        # pylint:disable=super-init-not-called
        def __init__(self, args):
            self.min = args["min"]
            self.max = args.get("max")

        definition = Directive(
            "len",
            ["ARGUMENT_DEFINITION", "INPUT_FIELD_DEFINITION"],
            [Argument("min", Int, default_value=0), Argument("max", Int)],
        )

        def visit_argument(self, arg):
            arg.type = LimitedLengthScalarType.wrap(
                arg.type, self.min, self.max
            )
            return arg

        def visit_input_field(self, field):
            field.type = LimitedLengthScalarType.wrap(
                field.type, self.min, self.max
            )
            return field

    schema = make_executable_schema(
        """
        type Query {
            foo (
                bar: BarInput
                foo: String @len(max: 4)
            ): String
        }

        input BarInput {
            baz: String @len(min: 3)
        }
        """,
        schema_directives={"len": LimitedLengthDirective},
    )

    assert schema.to_string() == dedent(
        """
        input BarInput {
            baz: LimitedLenthString_3_inf
        }

        type Query {
            foo(bar: BarInput, foo: LimitedLenthString_0_4): String
        }
        """
    )

    assert graphql(
        schema,
        '{ foo (foo: "abcdef") }',
        {},
        initial_value={"foo": lambda _, args, *r: args["foo"]},
    ).response() == {
        "errors": [
            {
                "locations": [{"column": 13, "line": 1}],
                "message": (
                    "Expected type LimitedLenthString_0_4, found "
                    '"abcdef" (LimitedLenthString_0_4 length must be '
                    "between 0 and 4 but was 6)"
                ),
            }
        ]
    }

    assert graphql(
        schema,
        '{ foo (foo: "abcd") }',
        {},
        initial_value={"foo": lambda _, args, *r: args["foo"]},
    ).response() == {"data": {"foo": "abcd"}}

    assert graphql(
        schema,
        '{ foo (bar: {baz: "abcd"}) }',
        {},
        initial_value={"foo": lambda _, args, *r: args["bar"]["baz"]},
    ).response() == {"data": {"foo": "abcd"}}

    assert graphql(
        schema,
        '{ foo (bar: {baz: "a"}) }',
        {},
        initial_value={"foo": lambda _, args, *r: args["bar"]["baz"]},
    ).response() == {
        "errors": [
            {
                "locations": [{"column": 19, "line": 1}],
                "message": (
                    'Expected type LimitedLenthString_3_inf, found "a" '
                    "(LimitedLenthString_3_inf length must be between 3 "
                    "and inf but was 1)"
                ),
            }
        ]
    }


def test_enum_value_directive():
    """ generating custom enum values """

    # These could be pre-loaded from a database or a config file dynamically
    VALUES = dict(
        [("RED", "#FF4136"), ("BLUE", "#0074D9"), ("GREEN", "#2ECC40")]
    )

    class CSSColorDirective(SchemaDirective):
        def visit_enum_value(self, enum_value):
            return EnumValue(
                enum_value.name,
                VALUES[enum_value.name],
                description=enum_value.description,
                deprecation_reason=enum_value.deprecation_reason,
            )

    schema = make_executable_schema(
        """
        directive @cssColor on ENUM_VALUE

        type Query {
            color: Color
        }

        enum Color {
            RED @cssColor
            BLUE @cssColor
            GREEN @cssColor
        }
        """,
        schema_directives={"cssColor": CSSColorDirective},
    )

    enum = schema.get_type("Color")
    for k, v in VALUES.items():
        assert enum.get_value(k) == v
        assert enum.get_name(v) == k


def test_enum_type_directive():
    """ generating custom enums """

    # These could be pre-loaded from a database or a config file dynamically
    VALUES = [("RED", "#FF4136"), ("BLUE", "#0074D9"), ("GREEN", "#2ECC40")]

    class GeneratedEnum(SchemaDirective):
        def visit_enum(self, enum):
            return EnumType(
                enum.name,
                VALUES,
                description=enum.description,
                nodes=enum.nodes,
            )

    schema = make_executable_schema(
        """
        directive @generated on ENUM

        type Query {
            color: Color
        }

        enum Color @generated { _empty }
        """,
        schema_directives={"generated": GeneratedEnum},
    )

    assert schema.to_string() == dedent(
        """
        directive @generated on ENUM

        enum Color {
            RED
            BLUE
            GREEN
        }

        type Query {
            color: Color
        }
        """
    )

    enum = schema.get_type("Color")
    values_dict = dict(VALUES)
    for k, v in values_dict.items():
        assert enum.get_value(k) == v
        assert enum.get_name(v) == k


def test_schema_extension_duplicate_directive():
    class OnSchema(SchemaDirective):
        def visit_schema(self, schema):
            pass

    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            directive @onSchema on SCHEMA

            type Foo { foo: String }
            type Bar { bar: String }

            schema @onSchema {
                query: Foo
            }

            extend schema @onSchema
            """,
            schema_directives={"onSchema": OnSchema},
        )

    assert exc_info.value.to_dict() == {
        "locations": [{"column": 27, "line": 11}],
        "message": 'Directive "@onSchema" already applied',
    }



# -*- coding: utf-8 -*-
""" Lexer tests.
Ported from: https://github.com/graphql/graphql-js/blob/master/src/language/__tests__/lexer-test.js  # noqa: E501
on revision 8d1ae25de444a9b544a1fdc98e138ae91b77057c.
"""
# Test naming is not very 'pythonic' as I tried to keep a close match with the
# original ones for easy reference. Others were kept while they were specific
# to the implementation and are kept as skipped.

import pytest

from py_gql.exc import (
    InvalidCharacter,
    InvalidEscapeSequence,
    NonTerminatedString,
    UnexpectedCharacter,
    UnexpectedEOF,
)
from py_gql.lang import Lexer, token


def lex_one(source):
    lexer = Lexer(source)
    assert type(next(lexer)) == token.SOF
    return next(lexer)


def test_it_disallows_uncommon_control_characters():
    with pytest.raises(InvalidCharacter) as exc_info:
        lex_one(u"\u0007")
    assert exc_info.value.position == 1


def test_it_accepts_bom_header():
    assert lex_one(u"\uFEFF foo") == token.Name(2, 5, "foo")


def test_it_accepts_binary_type():
    assert lex_one(b"foo") == token.Name(0, 3, "foo")


def test_it_skips_whitespace_and_comments_1():
    assert (
        lex_one(
            u"""

    foo


    """
        )
        == token.Name(6, 9, "foo")
    )


def test_it_skips_whitespace_and_comments_2():
    assert (
        lex_one(
            u"""
    #comment
    foo#comment
    """
        )
        == token.Name(18, 21, "foo")
    )


def test_it_skips_whitespace_and_comments_3():
    assert lex_one(u",,,foo,,,") == token.Name(3, 6, "foo")


def test_errors_respect_whitespace():
    with pytest.raises(UnexpectedCharacter) as exc_info:
        lex_one(
            u"""

                ?

            """
        )

    assert str(exc_info.value) == (
        'Unexpected character "?" (3:17):\n'
        "  1:\n"
        "  2:\n"
        "  3:                ?\n"
        "                    ^\n"
        "  4:\n"
        "  5:            "
    )


@pytest.mark.parametrize(
    "value,expected",
    [
        (u'"simple"', token.String(0, 8, "simple")),
        (u'" white space "', token.String(0, 15, " white space ")),
        (u'"quote \\""', token.String(0, 10, 'quote "')),
        (
            u'"escaped \\n\\r\\b\\t\\f"',
            token.String(0, 20, "escaped \n\r\b\t\f"),
        ),
        (u'"slashes \\\\ \\/"', token.String(0, 15, "slashes \\ /")),
        (
            u'"unicode \\u1234\\u5678\\u90AB\\uCDEF"',
            token.String(0, 34, u"unicode \u1234\u5678\u90AB\uCDEF"),
        ),
    ],
)
def test_it_lexes_strings(value, expected):
    assert lex_one(value) == expected


@pytest.mark.parametrize(
    "value, err_cls, expected_positon",
    [
        (u'"', NonTerminatedString, 1),
        (u'"no end quote', NonTerminatedString, 13),
        (u"'single quotes'", UnexpectedCharacter, 0),
        (u'"contains unescaped \u0007 control char"', InvalidCharacter, 20),
        (u'"null-byte is not \u0000 end of file"', InvalidCharacter, 18),
        (u'"multi\nline"', NonTerminatedString, 6),
        (u'"multi\rline"', NonTerminatedString, 6),
        (u'"bad \\z esc"', InvalidEscapeSequence, 6),
        (u'"bad \\x esc"', InvalidEscapeSequence, 6),
        (u'"bad \\u1 esc"', InvalidEscapeSequence, 6),
        (u'"bad \\u0XX1 esc"', InvalidEscapeSequence, 6),
        (u'"bad \\uXXXX esc"', InvalidEscapeSequence, 6),
        (u'"bad \\uFXXX esc"', InvalidEscapeSequence, 6),
        (u'"bad \\uXXXF esc"', InvalidEscapeSequence, 6),
    ],
)
def test_it_lex_reports_useful_string_errors(value, err_cls, expected_positon):
    with pytest.raises(err_cls) as exc_info:
        lex_one(value)
    assert exc_info.value.position == expected_positon


@pytest.mark.parametrize(
    "value, expected",
    [
        (u'"""simple"""', token.BlockString(0, 12, "simple")),
        (u'""" white space """', token.BlockString(0, 19, " white space ")),
        (
            u'"""contains " quote"""',
            token.BlockString(0, 22, 'contains " quote'),
        ),
        (
            u'"""contains \\""" triplequote"""',
            token.BlockString(0, 31, 'contains """ triplequote'),
        ),
        (u'"""multi\nline"""', token.BlockString(0, 16, "multi\nline")),
        (
            u'"""multi\rline\r\nnormalized"""',
            token.BlockString(0, 28, "multi\nline\nnormalized"),
        ),
        (
            u'"""unescaped \\n\\r\\b\\t\\f\\u1234"""',
            token.BlockString(0, 32, "unescaped \\n\\r\\b\\t\\f\\u1234"),
        ),
        (
            u'"""slashes \\\\ \\/"""',
            token.BlockString(0, 19, "slashes \\\\ \\/"),
        ),
        (
            u'''"""

        spans
          multiple
            lines

        """''',
            token.BlockString(0, 68, "spans\n  multiple\n    lines"),
        ),
    ],
)
def test_it_lexes_block_strings(value, expected):
    assert lex_one(value) == expected


@pytest.mark.parametrize(
    "value, err_cls, expected_positon",
    [
        (u'"""', NonTerminatedString, 3),
        (u'"""no end quote', NonTerminatedString, 15),
        (u'"""contains unescaped \u0007 control char"""', InvalidCharacter, 22),
        (u'"""null-byte is not \u0000 end of file"""', InvalidCharacter, 20),
    ],
)
def test_it_lex_reports_useful_block_string_errors(
    value, err_cls, expected_positon
):
    with pytest.raises(err_cls) as exc_info:
        lex_one(value)
    assert exc_info.value.position == expected_positon


@pytest.mark.parametrize(
    "string, expected",
    [
        (u"4", token.Integer(0, 1, "4")),
        (u"4.123", token.Float(0, 5, "4.123")),
        (u"-4", token.Integer(0, 2, "-4")),
        (u"9", token.Integer(0, 1, "9")),
        (u"0", token.Integer(0, 1, "0")),
        (u"-4.123", token.Float(0, 6, "-4.123")),
        (u"0.123", token.Float(0, 5, "0.123")),
        (u"123e4", token.Float(0, 5, "123e4")),
        (u"123E4", token.Float(0, 5, "123E4")),
        (u"123e-4", token.Float(0, 6, "123e-4")),
        (u"123e+4", token.Float(0, 6, "123e+4")),
        (u"-123e4", token.Float(0, 6, "-123e4")),
        (u"-123E4", token.Float(0, 6, "-123E4")),
        (u"-123e-4", token.Float(0, 7, "-123e-4")),
        (u"-123e+4", token.Float(0, 7, "-123e+4")),
        (u"-1.123e4567", token.Float(0, 11, "-1.123e4567")),
    ],
)
def test_it_lexes_numbers(string, expected):
    assert lex_one(string) == expected


@pytest.mark.parametrize(
    "value, err_cls, expected_positon",
    [
        (u"00", UnexpectedCharacter, 1),
        (u"+1", UnexpectedCharacter, 0),
        (u"1.", UnexpectedEOF, 2),
        (u"1.e1", UnexpectedCharacter, 2),
        (u".123", UnexpectedCharacter, 2),
        (u"1.A", UnexpectedCharacter, 2),
        (u"-A", UnexpectedCharacter, 1),
        (u"1.0e", UnexpectedEOF, 4),
        (u"1.0eA", UnexpectedCharacter, 4),
    ],
)
def test_it_lex_reports_useful_number_errors(value, err_cls, expected_positon):
    with pytest.raises(err_cls) as exc_info:
        lex_one(value)
    assert exc_info.value.position == expected_positon


@pytest.mark.parametrize(
    "string, expected",
    [
        (u"!", token.ExclamationMark(0, 1)),
        (u"$", token.Dollar(0, 1)),
        (u"(", token.ParenOpen(0, 1)),
        (u")", token.ParenClose(0, 1)),
        (u"[", token.BracketOpen(0, 1)),
        (u"]", token.BracketClose(0, 1)),
        (u"{", token.CurlyOpen(0, 1)),
        (u"}", token.CurlyClose(0, 1)),
        (u":", token.Colon(0, 1)),
        (u"=", token.Equals(0, 1)),
        (u"@", token.At(0, 1)),
        (u"|", token.Pipe(0, 1)),
        (u"&", token.Ampersand(0, 1)),
        (u"...", token.Ellipsis_(0, 3)),
    ],
)
def test_it_lexes_punctuation(string, expected):
    assert lex_one(string) == expected


@pytest.mark.parametrize(
    "value, err_cls, pos",
    [
        (u"..", UnexpectedEOF, 2),
        (u"?", UnexpectedCharacter, 0),
        (u"\u203B", UnexpectedCharacter, 0),
        (u"\u200b", UnexpectedCharacter, 0),
    ],
)
def test_it_lex_reports_useful_unknown_character_error(value, err_cls, pos):
    with pytest.raises(err_cls) as exc_info:
        lex_one(value)
    assert exc_info.value.position == pos


def test_it_lexes_multiple_tokens():
    assert (
        list(
            Lexer(
                u"""
    query {
        Node (search: "foo") {
            id
            name
        }
    }
    """
            )
        )
        == [
            token.SOF(0, 0),
            token.Name(5, 10, "query"),
            token.CurlyOpen(11, 12),
            token.Name(21, 25, "Node"),
            token.ParenOpen(26, 27),
            token.Name(27, 33, "search"),
            token.Colon(33, 34),
            token.String(35, 40, "foo"),
            token.ParenClose(40, 41),
            token.CurlyOpen(42, 43),
            token.Name(56, 58, "id"),
            token.Name(71, 75, "name"),
            token.CurlyClose(84, 85),
            token.CurlyClose(90, 91),
            token.EOF(96, 96),
        ]
    )


def test_kitchen_sink(fixture_file):
    source = fixture_file("kitchen-sink.graphql")
    assert list(Lexer(source))



# -*- coding: utf-8 -*-

from concurrent import futures as _f

import pytest

from py_gql.execution import _concurrency


def test_all_wrapper_futures_ok():
    futures = [_f.Future() for _ in range(10)]

    wrapper = _concurrency.all_(futures)
    for i, f in enumerate(futures):
        f.set_result(i)

    assert wrapper.result() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_all_wrapper_futures_mixed():
    futures = []

    for x in range(10):
        if x % 2:
            f = _f.Future()
            futures.append(f)
        else:
            futures.append(x)

    wrapper = _concurrency.all_(futures)
    for i, f in enumerate(futures):
        if i % 2:
            f.set_result(i)

    assert wrapper.result() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_all_wrapper_partially_completed():
    futures = [_f.Future() for _ in range(10)]

    for i, f in enumerate(futures):
        if i % 2:
            f.set_result(i)

    wrapper = _concurrency.all_(futures)

    for i, f in enumerate(futures):
        if not i % 2:
            f.set_result(i)

    assert wrapper.result() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_all_wrapper_one_future_fails():
    futures = [_f.Future() for _ in range(10)]
    for i, f in enumerate(futures):
        if i == 9:
            f.set_exception(ValueError("I don't like 9"))
        else:
            f.set_result(i)

    wrapper = _concurrency.all_(futures)

    with pytest.raises(ValueError) as exc_info:
        wrapper.result()

    assert str(exc_info.value) == "I don't like 9"


def test_chain_wrapper_sync_steps():
    f = _f.Future()
    c = _concurrency.chain(f, lambda x: x + 1, lambda x: x * 2)
    f.set_result(1)
    assert c.result() == 4


def test_chain_wrapper_sync_failure():
    f = _f.Future()

    def _bad_step(x):
        raise ValueError(x)

    c = _concurrency.chain(f, _bad_step)
    f.set_result(1)

    with pytest.raises(ValueError) as exc_info:
        c.result(timeout=1)

    assert str(exc_info.value) == "1"


def test_chain_wrapper_async_step():
    with _f.ThreadPoolExecutor(max_workers=1) as executor:

        def sync_step(x):
            return x * 2

        def async_step(x):
            return executor.submit(sync_step, x)

        c = _concurrency.chain(1, lambda x: x + 1, async_step, lambda x: x + 1)
        assert c.result() == 5


def test_chain_wrapper_async_failure():
    with _f.ThreadPoolExecutor(max_workers=1) as executor:

        def sync_step(x):
            raise ValueError(x)

        def async_step(x):
            return executor.submit(sync_step, x)

        c = _concurrency.chain(1, lambda x: x + 1, async_step, lambda x: x + 1)
        with pytest.raises(ValueError) as exc_info:
            c.result()

        assert str(exc_info.value) == "2"


def test_serial_ok_sync():

    called = []

    def make_step(x):
        def _step():
            assert called == list(range(x))
            called.append(x)

        return _step

    assert _concurrency.serial([make_step(x) for x in range(10)]) is None
    assert called == list(range(10))


def test_serial_ok_async():

    called = []

    def make_step(x):
        def _step():
            assert called == list(range(x))
            f = _f.Future()
            f.set_result(x)
            called.append(x)
            return f

        return _step

    assert _concurrency.serial([make_step(x) for x in range(10)]).result()
    assert called == list(range(10))


def test_unwrap_ok():
    futures = [_f.Future() for _ in range(10)]
    futures[-1].set_result("final")
    for i, f in enumerate(futures[:-1]):
        f.set_result(futures[i + 1])

    assert _concurrency.unwrap(futures[0]).result() == "final"


def test_unwrap_intermediate_failure():
    futures = [_f.Future() for _ in range(10)]
    futures[-1].set_result("final")
    for i, f in enumerate(futures[:-1]):
        if i != 5:
            f.set_result(futures[i + 1])
        else:
            f.set_exception(ValueError(5))

    with pytest.raises(ValueError) as exc_info:
        _concurrency.unwrap(futures[0]).result()

    assert str(exc_info.value) == "5"


def test_unwrap_intermediate_cancel():
    futures = [_f.Future() for _ in range(10)]
    futures[-1].set_result("final")
    for i, f in enumerate(futures[:-1]):
        if i != 5:
            f.set_result(futures[i + 1])
        else:
            f.cancel()

    with pytest.raises(_f.CancelledError):
        _concurrency.unwrap(futures[0]).result()


def test_except_wrapper():
    f = _f.Future()
    e = _concurrency.except_(f)
    f.set_exception(ValueError("foo"))
    assert e.result() is None


def test_except_wrapper_with_mapper():
    f = _f.Future()
    e = _concurrency.except_(f, map_=str)
    f.set_exception(ValueError("foo"))
    assert e.result() == "foo"


def test_except_wrapper_correct_exc_cls():
    f = _f.Future()
    e = _concurrency.except_(f, (ValueError,))
    f.set_exception(ValueError("foo"))
    assert e.result() is None


def test_except_wrapper_incorrect_exc_cls():
    f = _f.Future()
    e = _concurrency.except_(f, (TypeError,))
    f.set_exception(ValueError("foo"))

    with pytest.raises(ValueError):
        e.result()


def test_DummyFuture_not_cancellable():
    f = _concurrency.DummyFuture()
    assert not f.cancel()


def test_DummyFuture_never_cancelled():
    f = _concurrency.DummyFuture()
    f.cancel()
    assert not f.cancelled()


def test_DummyFuture_defaults_to_running():
    f = _concurrency.DummyFuture()
    assert f.running()


def test_DummyFuture_not_running_after_result():
    f = _concurrency.DummyFuture()
    f.set_result("foo")
    assert not f.running()


def test_DummyFuture_not_running_after_exception():
    f = _concurrency.DummyFuture()
    f.set_exception(ValueError("1"))
    assert not f.running()


def test_DummyFuture_raises_on_blocking_call_1():
    f = _concurrency.DummyFuture()
    with pytest.raises(RuntimeError):
        f.result()


def test_DummyFuture_raises_on_blocking_call_2():
    f = _concurrency.DummyFuture()
    with pytest.raises(RuntimeError):
        f.exception()


def test_DummyFuture_calls_callbacks_correctly(mocker):
    cb1, cb2, cb3 = mocker.Mock(), mocker.Mock(), mocker.Mock()
    f = _concurrency.DummyFuture()

    f.add_done_callback(cb1)
    f.add_done_callback(cb2)

    f.set_result("foo")

    cb1.assert_called_with(f)
    cb2.assert_called_with(f)
    cb3.assert_not_called()

    f.add_done_callback(cb3)

    cb3.assert_called_with(f)


def test_DummyFuture_result_returns_when_done_with_result():
    f = _concurrency.DummyFuture()
    f.set_result("foo")
    assert f.result() == "foo"


def test_DummyFuture_result_raises_when_done_with_exception():
    f = _concurrency.DummyFuture()
    f.set_exception(ValueError("foo"))
    with pytest.raises(ValueError):
        f.result()


def test_DummyFuture_exception_returns_when_done_with_exception():
    f = _concurrency.DummyFuture()
    err = ValueError("foo")
    f.set_exception(err)
    assert f.exception() == err


def test_DummyFuture_exception_returns_when_done_with_result():
    f = _concurrency.DummyFuture()
    f.set_result("foo")
    assert f.exception() is None



# -*- coding: utf-8 -*-
""" Parser tests.
Ported from: https://github.com/graphql/graphql-js/blob/master/src/language/__tests__/parser-test.js  # noqa: E501
on revision 8d1ae25de444a9b544a1fdc98e138ae91b77057c.
"""
# Test naming is not very 'pythonic' as I tried to keep a close match with the
# original ones for easy reference. Others were kept while they were specific
# to the implementation and are kept as skipped.

import pytest

from py_gql.exc import GraphQLSyntaxError, UnexpectedToken
from py_gql.lang import ast as _ast
from py_gql.lang.parser import parse, parse_type, parse_value


# Comparing dicts will result in better assertion diffs from pytest.
def assert_node_equal(ref, expected):
    assert _ast.node_to_dict(ref) == _ast.node_to_dict(expected)


def test_it_asserts_that_a_source_to_parse_was_provided():
    with pytest.raises(ValueError):
        parse(None)


@pytest.mark.parametrize(
    "value, error_cls, position, message",
    [
        (u"{", UnexpectedToken, 1, 'Expected Name but found "<EOF>"'),
        (
            u"\n{ ...MissingOn }\nfragment MissingOn Type",
            UnexpectedToken,
            37,
            'Expected "on" but found "Type"',
        ),
        (u"{ field: {} }", UnexpectedToken, 9, 'Expected Name but found "{"'),
        (
            u"notanoperation Foo { field }",
            UnexpectedToken,
            0,
            'Unexpected "notanoperation"',
        ),
        (u"...", UnexpectedToken, 0, 'Unexpected "..."'),
    ],
)
def test_it_provides_useful_errors(value, error_cls, position, message):
    with pytest.raises(error_cls) as exc_info:
        parse(value)

    if position is not None:
        assert exc_info.value.position == position
    if message is not None:
        assert exc_info.value.message == message


def test_it_parses_variable_inline_values():
    # assert doesn't raise
    parse(u"{ field(complex: { a: { b: [ $var ] } }) }")


def test_it_parses_constant_default_values():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(u"query Foo($x: Complex = { a: { b: [ $var ] } }) { field }")
    assert exc_info.value.position == 36
    assert exc_info.value.message == 'Unexpected "$"'


def test_it_does_not_accept_fragments_named_on():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(u"fragment on on on { on }")
    assert exc_info.value.position == 9
    assert exc_info.value.message == 'Unexpected "on"'


def test_it_does_not_accept_fragments_spread_of_on():
    with pytest.raises(UnexpectedToken) as exc_info:
        parse(u"{ ...on }")
    assert exc_info.value.position == 8
    assert exc_info.value.message == 'Expected Name but found "}"'


def test_it_parses_multi_bytes_characters():
    source = u"""
        # This comment has a \u0A0A multi-byte character.
        { field(arg: "Has a \u0A0A multi-byte character.") }
      """
    tree = parse(source, no_location=True)
    assert_node_equal(
        tree.definitions[0].selection_set.selections,
        [
            _ast.Field(
                name=_ast.Name(value="field"),
                arguments=[
                    _ast.Argument(
                        name=_ast.Name(value="arg"),
                        value=_ast.StringValue(
                            value=u"Has a \u0A0A multi-byte character."
                        ),
                    )
                ],
            )
        ],
    )


def test_it_parses_kitchen_sink(fixture_file):
    # assert doesn't raise
    source = fixture_file("kitchen-sink.graphql")
    assert parse(source, no_location=True)
    assert parse(source, no_location=True, allow_type_system=False)


@pytest.mark.parametrize(
    "keyword",
    [
        (u"on"),
        (u"fragment"),
        (u"query"),
        (u"mutation"),
        (u"subscription"),
        (u"true"),
        (u"false"),
    ],
)
def test_it_allows_non_keywords_anywhere_a_name_is_allowed(keyword):
    fragment_name = keyword if keyword != "on" else "a"
    assert parse(
        u"""
    query %(keyword)s {
        ... %(fragment_name)s
        ... on %(keyword)s { field }
    }
    fragment %(fragment_name)s on Type {
        %(keyword)s(%(keyword)s: $%(keyword)s)
            @%(keyword)s(%(keyword)s: %(keyword)s)
    }"""
        % dict(keyword=keyword, fragment_name=fragment_name)
    )


def test_it_parses_anonymous_mutation_operations():
    # assert doesn't raise
    assert parse(
        u"""
    mutation {
        mutationField
    }
    """,
        no_location=True,
    )


def test_it_parses_anonymous_subscription_operations():
    # assert doesn't raise
    assert parse(
        u"""
    subscription {
        subscriptionField
    }
    """,
        no_location=True,
    )


def test_it_parses_named_mutation_operations():
    assert parse(
        u"""
    mutation Foo {
        mutationField
    }
    """,
        no_location=True,
    )


def test_it_parses_named_subscription_operations():
    assert parse(
        u"""
    subscription Foo {
        subscriptionField
    }""",
        no_location=True,
    )


def test_it_creates_ast():
    assert_node_equal(
        parse(
            u"""{
  node(id: 4) {
    id,
    name
  }
}
"""
        ),
        _ast.Document(
            loc=(0, 41),
            definitions=[
                _ast.OperationDefinition(
                    loc=(0, 40),
                    operation="query",
                    name=None,
                    variable_definitions=[],
                    directives=[],
                    selection_set=_ast.SelectionSet(
                        loc=(0, 40),
                        selections=[
                            _ast.Field(
                                loc=(4, 38),
                                alias=None,
                                name=_ast.Name(loc=(4, 8), value="node"),
                                arguments=[
                                    _ast.Argument(
                                        loc=(9, 14),
                                        name=_ast.Name(loc=(9, 11), value="id"),
                                        value=_ast.IntValue(
                                            loc=(13, 14), value="4"
                                        ),
                                    )
                                ],
                                directives=[],
                                selection_set=_ast.SelectionSet(
                                    loc=(16, 38),
                                    selections=[
                                        _ast.Field(
                                            loc=(22, 24),
                                            alias=None,
                                            name=_ast.Name(
                                                loc=(22, 24), value="id"
                                            ),
                                            directives=[],
                                            arguments=[],
                                            selection_set=None,
                                        ),
                                        _ast.Field(
                                            loc=(30, 34),
                                            alias=None,
                                            name=_ast.Name(
                                                loc=(30, 34), value="name"
                                            ),
                                            directives=[],
                                            arguments=[],
                                            selection_set=None,
                                        ),
                                    ],
                                ),
                            )
                        ],
                    ),
                )
            ],
        ),
    )


def test_it_creates_ast_from_nameless_query_without_variables():
    body = u"""query {
  node {
    id
  }
}
"""
    assert_node_equal(
        parse(body),
        _ast.Document(
            loc=(0, 30),
            definitions=[
                _ast.OperationDefinition(
                    loc=(0, 29),
                    operation="query",
                    name=None,
                    variable_definitions=[],
                    directives=[],
                    selection_set=_ast.SelectionSet(
                        loc=(6, 29),
                        selections=[
                            _ast.Field(
                                loc=(10, 27),
                                alias=None,
                                name=_ast.Name(loc=(10, 14), value="node"),
                                arguments=[],
                                directives=[],
                                selection_set=_ast.SelectionSet(
                                    loc=(15, 27),
                                    selections=[
                                        _ast.Field(
                                            loc=(21, 23),
                                            alias=None,
                                            name=_ast.Name(
                                                loc=(21, 23), value="id"
                                            ),
                                            arguments=[],
                                            directives=[],
                                            selection_set=None,
                                        )
                                    ],
                                ),
                            )
                        ],
                    ),
                )
            ],
        ),
    )


def test_it_allows_parsing_without_source_location_information():
    assert parse(u"{ id }", no_location=True).loc is None


def test_it_experimental_allows_parsing_fragment_defined_variables():
    with pytest.raises(GraphQLSyntaxError):
        parse(u"fragment a($v: Boolean = false) on t { f(v: $v) }")

    assert parse(
        u"fragment a($v: Boolean = false) on t { f(v: $v) }",
        experimental_fragment_variables=True,
    )


def test_it_contains_references_to_source():
    doc = parse(u"{ id }")
    assert doc.source == "{ id }"


def test_parse_value_it_parses_null_value():
    assert_node_equal(parse_value(u"null"), _ast.NullValue(loc=(0, 4)))


def test_parse_value_it_parses_list_values():
    assert_node_equal(
        parse_value(u'[123 "abc"]'),
        _ast.ListValue(
            loc=(0, 11),
            values=[
                _ast.IntValue(loc=(1, 4), value="123"),
                _ast.StringValue(loc=(5, 10), value="abc"),
            ],
        ),
    )


def test_parse_value_it_parses_block_strings():
    assert_node_equal(
        parse_value(u'["""long""" "short"]'),
        _ast.ListValue(
            loc=(0, 20),
            values=[
                _ast.StringValue(loc=(1, 11), value="long", block=True),
                _ast.StringValue(loc=(12, 19), value="short"),
            ],
        ),
    )


def test_parse_type_it_parses_well_known_types():
    assert_node_equal(
        parse_type(u"String"),
        _ast.NamedType(loc=(0, 6), name=_ast.Name(loc=(0, 6), value="String")),
    )


def test_parse_type_it_parses_custom_types():
    assert_node_equal(
        parse_type(u"MyType"),
        _ast.NamedType(loc=(0, 6), name=_ast.Name(loc=(0, 6), value="MyType")),
    )


def test_parse_type_it_parses_list_types():
    assert_node_equal(
        parse_type(u"[MyType]"),
        _ast.ListType(
            loc=(0, 8),
            type=_ast.NamedType(
                loc=(1, 7), name=_ast.Name(loc=(1, 7), value="MyType")
            ),
        ),
    )


def test_parse_type_it_parses_non_null_types():
    assert_node_equal(
        parse_type(u"MyType!"),
        _ast.NonNullType(
            loc=(0, 7),
            type=_ast.NamedType(
                loc=(0, 6), name=_ast.Name(loc=(0, 6), value="MyType")
            ),
        ),
    )


def test_parse_type_it_parses_nested_types():
    assert_node_equal(
        parse_type(u"[MyType!]"),
        _ast.ListType(
            loc=(0, 9),
            type=_ast.NonNullType(
                loc=(1, 8),
                type=_ast.NamedType(
                    loc=(1, 7), name=_ast.Name(loc=(1, 7), value="MyType")
                ),
            ),
        ),
    )


# Extra tests not in the reference
# Mostly discovered during testing other parts of the library
# ------------------------------------------------------------------------------


def test_parse_type_it_parses_nested_types_2():
    assert_node_equal(
        parse_type(u"[MyType!]!"),
        _ast.NonNullType(
            loc=(0, 10),
            type=_ast.ListType(
                loc=(0, 9),
                type=_ast.NonNullType(
                    loc=(1, 8),
                    type=_ast.NamedType(
                        loc=(1, 7), name=_ast.Name(loc=(1, 7), value="MyType")
                    ),
                ),
            ),
        ),
    )


def test_it_parses_inline_fragment_without_type():
    assert_node_equal(
        parse(
            """
    fragment validFragment on Pet {
        ... {
            name
        }
    }
    """
        ),
        _ast.Document(
            loc=(0, 88),
            definitions=[
                _ast.FragmentDefinition(
                    loc=(5, 83),
                    name=_ast.Name(loc=(14, 27), value="validFragment"),
                    type_condition=_ast.NamedType(
                        loc=(31, 34), name=_ast.Name(loc=(31, 34), value="Pet")
                    ),
                    variable_definitions=None,
                    selection_set=_ast.SelectionSet(
                        loc=(35, 83),
                        selections=[
                            _ast.InlineFragment(
                                loc=(45, 77),
                                selection_set=_ast.SelectionSet(
                                    loc=(49, 77),
                                    selections=[
                                        _ast.Field(
                                            loc=(63, 67),
                                            name=_ast.Name(
                                                loc=(63, 67), value="name"
                                            ),
                                        )
                                    ],
                                ),
                            )
                        ],
                    ),
                )
            ],
        ),
    )



# -*- coding: utf-8 -*-

import json

from six.moves import map

from . import ast as _ast


class ASTPrinter(object):
    """ String formatter for ast node.

    Args:
        indent (Union[str, int]): Indent character or number of spaces

        include_descriptions (bool):
            If ``True`` include descriptions as leading block strings in the
            output. Only relevant for SDL nodes.

        description_format ("comments"|"block"):
            Control how descriptions are formatted. ``"comments"`` is the
            old standard and will be compatible with most GraphQL parsers
            while ``"block"`` is part of the most recent specification and
            includes descriptions as block strings that can be extracted
            according to the specification.
    """

    __slots__ = ("indent", "include_descriptions", "description_format")

    def __init__(
        self, indent=4, include_descriptions=True, description_format="block"
    ):
        self.include_descriptions = include_descriptions
        self.description_format = description_format
        if isinstance(indent, int):
            self.indent = indent * " "
        else:
            self.indent = indent

    def __call__(self, node):
        """ Converts an AST into a string, using a set of reasonable
        formatting rules.

        Args:
            node (py_gql.lang.ast.Node): Input node

        Returns:
            str: Formatted value for the provided node
        """
        if node is None:
            return ""

        kind = type(node)

        try:
            printer = {
                _ast.Name: self.print_name,
                _ast.Variable: self.print_variable,
                _ast.Document: self.print_document,
                _ast.OperationDefinition: self.print_operation_definition,
                _ast.VariableDefinition: self.print_variable_definition,
                _ast.SelectionSet: self.print_selection_set,
                _ast.Field: self.print_field,
                _ast.Argument: self.print_argument,
                _ast.FragmentSpread: self.print_fragment_spread,
                _ast.InlineFragment: self.print_inline_fragment,
                _ast.IntValue: self.print_int_value,
                _ast.FloatValue: self.print_float_value,
                _ast.EnumValue: self.print_enum_value,
                _ast.BooleanValue: self.print_boolean_value,
                _ast.NullValue: self.print_null_value,
                _ast.StringValue: self.print_string_value,
                _ast.ListValue: self.print_list_value,
                _ast.ObjectValue: self.print_object_value,
                _ast.ObjectField: self.print_object_field,
                _ast.Directive: self.print_directive,
                _ast.NamedType: self.print_named_type,
                _ast.ListType: self.print_list_type,
                _ast.NonNullType: self.print_non_null_type,
                _ast.FragmentDefinition: self.print_fragment_definition,
                _ast.SchemaDefinition: self.print_schema_definition,
                _ast.SchemaExtension: self.print_schema_extension,
                _ast.OperationTypeDefinition: self.print_operation_type_definition,
                _ast.ScalarTypeDefinition: self.print_scalar_type_definition,
                _ast.ScalarTypeExtension: self.print_scalar_type_extension,
                _ast.ObjectTypeDefinition: self.print_object_type_definition,
                _ast.ObjectTypeExtension: self.print_object_type_extension,
                _ast.FieldDefinition: self.print_field_definition,
                _ast.InputValueDefinition: self.print_input_value_definition,
                _ast.InterfaceTypeDefinition: self.print_interface_type_definition,
                _ast.InterfaceTypeExtension: self.print_interface_type_extension,
                _ast.UnionTypeDefinition: self.print_union_type_definition,
                _ast.UnionTypeExtension: self.print_union_type_extension,
                _ast.EnumTypeDefinition: self.print_enum_type_definition,
                _ast.EnumTypeExtension: self.print_enum_type_extension,
                _ast.EnumValueDefinition: self.print_enum_value_definition,
                _ast.InputObjectTypeDefinition: self.print_input_object_type_definition,
                _ast.InputObjectTypeExtension: self.print_input_object_type_extension,
                _ast.DirectiveDefinition: self.print_directive_definition,
            }[kind]
        except KeyError:
            raise TypeError(kind)

        return printer(node)

    def print_name(self, node):
        # type: (_ast.Name) -> str
        return node.value

    def print_variable(self, node):
        # type: (_ast.Variable) -> str
        return "$%s" % node.name.value

    def print_document(self, node):
        # type: (_ast.Document) -> str
        return _join(map(self, node.definitions), "\n\n") + "\n"

    def print_operation_definition(self, node):
        # type: (_ast.OperationDefinition) -> str
        op = node.operation
        name = node.name.value if node.name else ""
        var_defs = self.print_variable_definitions(node)
        directives = self.print_directives(node)
        selection_set = self._selection_set(node)
        use_short_form = (
            not name
            and not directives
            and not var_defs
            and (op == "query" or not op)
        )
        if use_short_form:
            return selection_set
        return _join(
            [op, _join([name, var_defs]), directives, selection_set], " "
        )

    def print_variable_definition(self, node):
        # type: (_ast.VariableDefinition) -> str
        return "%s: %s%s" % (
            self.print_variable(node.variable),
            self(node.type),
            _wrap(" = ", self(node.default_value)),
        )

    def print_variable_definitions(self, node):
        return _wrap(
            "(",
            _join(
                map(self.print_variable_definition, node.variable_definitions),
                ", ",
            ),
            ")",
        )

    def _selection_set(self, node):
        return (
            self.print_selection_set(node.selection_set)
            if node.selection_set
            else ""
        )

    def print_selection_set(self, node):
        # type: (_ast.SelectionSet) -> str
        return _block(map(self, node.selections), self.indent)

    def print_field(self, node):
        # type: (_ast.Field) -> str
        if node.alias:
            lead = _join([_wrap("", node.alias.value, ": "), node.name.value])
        else:
            lead = node.name.value

        return _join(
            [
                _join([lead, self.print_arguments(node)]),
                self.print_directives(node),
                self._selection_set(node),
            ],
            " ",
        )

    def print_arguments(self, node):
        return _wrap(
            "(", _join(map(self.print_argument, node.arguments), ", "), ")"
        )

    def print_argument(self, node):
        # type: (_ast.Argument) -> str
        return "%s: %s" % (node.name.value, self(node.value))

    def print_fragment_spread(self, node):
        # type: (_ast.FragmentSpread) -> str
        return "...%s%s" % (
            node.name.value,
            _wrap(" ", self.print_directives(node)),
        )

    def print_inline_fragment(self, node):
        # type: (_ast.InlineFragment) -> str
        return _join(
            [
                "...",
                _wrap("on ", self(node.type_condition)),
                self.print_directives(node),
                self._selection_set(node),
            ],
            " ",
        )

    def print_int_value(self, node):
        # type: (_ast.IntValue) -> str
        return node.value

    def print_float_value(self, node):
        # type: (_ast.FloatValue) -> str
        return node.value

    def print_enum_value(self, node):
        # type: (_ast.EnumValue) -> str
        return node.value

    def print_boolean_value(self, node):
        # type: (_ast.BooleanValue) -> str
        return str(node.value).lower()

    def print_null_value(self, node):
        # type: (_ast.NullValue) -> str
        return "null"

    def print_string_value(self, node):
        # type: (_ast.StringValue) -> str
        value = node.value
        return (
            _block_string(value, self.indent)
            if node.block
            else json.dumps(value)
        )

    def print_list_value(self, node):
        # type: (_ast.ListValue) -> str
        return "[%s]" % _join(map(self, node.values), ", ")

    def print_object_value(self, node):
        # type: (_ast.ObjectValue) -> str
        return "{%s}" % _join(map(self.print_object_field, node.fields), ", ")

    def print_object_field(self, node):
        # type: (_ast.ObjectField) -> str
        return "%s: %s" % (node.name.value, self(node.value))

    def print_directives(self, node):
        return _join(map(self.print_directive, node.directives), " ")

    def print_directive(self, node):
        # type: (_ast.Directive) -> str
        return "@%s%s" % (node.name.value, self.print_arguments(node))

    def print_named_type(self, node):
        # type: (_ast.NamedType) -> str
        return node.name.value

    def print_list_type(self, node):
        # type: (_ast.ListType) -> str
        return "[%s]" % self(node.type)

    def print_non_null_type(self, node):
        # type: (_ast.NonNullType) -> str
        return "%s!" % self(node.type)

    # NOTE: fragment variable definitions are experimental and may be
    # changed or removed in the future.

    def print_fragment_definition(self, node):
        # type: (_ast.FragmentDefinition) -> str
        return "fragment %s%s on %s %s%s" % (
            node.name.value,
            self.print_variable_definitions(node),
            self(node.type_condition),
            self.print_directives(node),
            self._selection_set(node),
        )

    def print_schema_definition(self, node):
        # type: (_ast.SchemaDefinition) -> str
        return _join(
            [
                "schema",
                self.print_directives(node),
                _block(map(self, node.operation_types), self.indent),
            ],
            " ",
        )

    def print_schema_extension(self, node):
        # type: (_ast.SchemaExtension) -> str
        return _join(
            [
                "extend schema",
                self.print_directives(node),
                _block(map(self, node.operation_types), self.indent),
            ],
            " ",
        )

    def print_operation_type_definition(self, node):
        # type: (_ast.OperationTypeDefinition) -> str
        return "%s: %s" % (node.operation, self(node.type))

    def print_scalar_type_definition(self, node):
        # type: (_ast.ScalarTypeDefinition) -> str
        return self._with_desc(
            _join(
                ["scalar", node.name.value, self.print_directives(node)], " "
            ),
            node.description,
        )

    def print_scalar_type_extension(self, node):
        # type: (_ast.ScalarTypeExtension) -> str
        return _join(
            ["extend scalar", node.name.value, self.print_directives(node)], " "
        )

    def print_object_type_definition(self, node):
        # type: (_ast.ObjectTypeDefinition) -> str
        return self._with_desc(
            _join(
                [
                    "type",
                    node.name.value,
                    _wrap(
                        "implements ", _join(map(self, node.interfaces), " & ")
                    ),
                    self.print_directives(node),
                    _block(map(self, node.fields), self.indent),
                ],
                " ",
            ),
            node.description,
        )

    def print_object_type_extension(self, node):
        # type: (_ast.ObjectTypeExtension) -> str
        return _join(
            [
                "extend type",
                node.name.value,
                _wrap("implements ", _join(map(self, node.interfaces), " & ")),
                self.print_directives(node),
                _block(map(self, node.fields), self.indent),
            ],
            " ",
        )

    def print_field_definition(self, node):
        # type: (_ast.FieldDefinition) -> str
        return _join(
            [
                node.name.value,
                self.print_argument_definitions(node),
                ": ",
                self(node.type),
                _wrap(" ", self.print_directives(node)),
            ]
        )

    def print_input_value_definition(self, node):
        # type: (_ast.InputValueDefinition) -> str
        return _join(
            [
                _join([node.name.value, ": ", self(node.type)]),
                _wrap(" = ", self(node.default_value)),
                _wrap(" ", self.print_directives(node)),
            ]
        )

    def print_interface_type_definition(self, node):
        # type: (_ast.InterfaceTypeDefinition) -> str
        return self._with_desc(
            _join(
                [
                    "interface",
                    node.name.value,
                    self.print_directives(node),
                    _block(map(self, node.fields), self.indent),
                ],
                " ",
            ),
            node.description,
        )

    def print_interface_type_extension(self, node):
        # type: (_ast.InterfaceTypeExtension) -> str
        return _join(
            [
                "extend interface",
                node.name.value,
                self.print_directives(node),
                _block(map(self, node.fields), self.indent),
            ],
            " ",
        )

    def print_union_type_definition(self, node):
        # type: (_ast.UnionTypeDefinition) -> str
        return self._with_desc(
            _join(
                [
                    "union",
                    node.name.value,
                    self.print_directives(node),
                    _wrap("= ", _join(map(self, node.types), " | ")),
                ],
                " ",
            ),
            node.description,
        )

    def print_union_type_extension(self, node):
        # type: (_ast.UnionTypeExtension) -> str
        return _join(
            [
                "extend union",
                node.name.value,
                self.print_directives(node),
                _wrap("= ", _join(map(self, node.types), " | ")),
            ],
            " ",
        )

    def print_enum_type_definition(self, node):
        # type: (_ast.EnumTypeDefinition) -> str
        return self._with_desc(
            _join(
                [
                    "enum",
                    node.name.value,
                    self.print_directives(node),
                    _block(map(self, node.values), self.indent),
                ],
                " ",
            ),
            node.description,
        )

    def print_enum_type_extension(self, node):
        # type: (_ast.EnumTypeExtension) -> str
        return _join(
            [
                "extend enum",
                node.name.value,
                self.print_directives(node),
                _block(map(self, node.values), self.indent),
            ],
            " ",
        )

    def print_enum_value_definition(self, node):
        # type: (_ast.EnumValueDefinition) -> str
        return _join([node.name.value, self.print_directives(node)], " ")

    def print_input_object_type_definition(self, node):
        # type: (_ast.InputObjectTypeDefinition) -> str
        return self._with_desc(
            _join(
                [
                    "input",
                    node.name.value,
                    self.print_directives(node),
                    _block(map(self, node.fields), self.indent),
                ],
                " ",
            ),
            node.description,
        )

    def print_input_object_type_extension(self, node):
        # type: (_ast.InputObjectTypeExtension) -> str
        return _join(
            [
                "extend input",
                node.name.value,
                self.print_directives(node),
                _block(map(self, node.fields), self.indent),
            ],
            " ",
        )

    def print_directive_definition(self, node):
        # type: (_ast.DirectiveDefinition) -> str
        return self._with_desc(
            _join(
                [
                    "directive @",
                    node.name.value,
                    self.print_argument_definitions(node),
                    " on ",
                    _join(map(self, node.locations), " | "),
                ]
            ),
            node.description,
        )

    def print_argument_definitions(self, node):
        args = list(map(self, node.arguments))
        if not any(("\n" in a for a in args)):
            return _wrap("(", _join(args, ", "), ")")
        else:
            return _wrap("(\n", _indent(_join(args, "\n"), self.indent), "\n)")

    def _with_desc(self, formatted, desc):
        # type: (str, _ast.StringValue) -> str
        if desc is None or not self.include_descriptions:
            return formatted

        if self.description_format == "block":
            desc_str = _block_string(desc.value, self.indent, True)
        elif self.description_format == "comments":
            desc_str = "\n".join(
                ("# " + line for line in desc.value.split("\n"))
            )
        else:
            raise ValueError(
                "Invalid description format %s" % self.description_format
            )

        return _join([desc_str, formatted], "\n")


def _wrap(start, maybe_string, end=None):
    return ("%s%s%s" % (start, maybe_string, end or "")) if maybe_string else ""


def _join(entries, separator=""):
    if entries is None:
        entries = []
    return separator.join([x for x in entries if x])


def _indent(maybe_string, indent):
    return maybe_string and (indent + maybe_string.replace("\n", "\n  "))


def _block(iterator, indent):
    arr = list(iterator)
    if not arr:
        return ""
    return "{\n%s\n}" % _join(map(lambda s: _indent(s, indent), arr), "\n")


# Print a block string in the indented block form by adding a leading and
# trailing blank line. However, if a block string starts with whitespace and
# is a single-line, adding a leading blank line would strip that whitespace.
def _block_string(value, indent, is_description=False):
    escaped = value.replace('"""', '\\"""')
    if (value[0] == " " or value[0] == "\t") and "\n" not in value:
        if escaped.endswith('"'):
            escaped = escaped + "\n"
        return '"""%s"""' % escaped
    return '"""\n%s\n"""' % (
        escaped if is_description else _indent(escaped, indent)
    )


# Default AST formatter matching graphql-js's implementation
print_ast = ASTPrinter(
    indent=2, include_descriptions=True, description_format="block"
)



# -*- coding: utf-8 -*-
""" GaphQL types related to introspection queries that
should be present in all spec compliant servers. """
from __future__ import unicode_literals

import json

import six

from .scalars import Boolean, String
from .types import (
    Argument,
    EnumType,
    EnumValue,
    Field,
    InputObjectType,
    InterfaceType,
    ListType,
    NonNullType,
    ObjectType,
    ScalarType,
    UnionType,
    WrappingType,
    is_abstract_type,
)

__Schema__ = ObjectType(
    "__Schema",
    description=(
        "A GraphQL Schema defines the capabilities of a GraphQL server. "
        "It exposes all available types and directives on the server, "
        "as well as the entry points for query, mutation, "
        "and subscription operations."
    ),
    fields=lambda: [
        Field(
            "types",
            NonNullType(ListType(NonNullType(__Type__))),
            description="A list of all types supported by this server.",
            # Sort as the output should not be determined by the way the schema
            # is built and this is the most logical order.
            resolve=lambda schema, *a, **kw: list(
                sorted(schema.types.values(), key=lambda t: t.name)
            ),
        ),
        Field(
            "queryType",
            NonNullType(__Type__),
            description="The type that query operations will be rooted at.",
            resolve=lambda schema, *a, **kw: schema.query_type,
        ),
        Field(
            "mutationType",
            __Type__,
            description=(
                "If this server supports mutation, the type that mutation "
                "operations will be rooted at."
            ),
            resolve=lambda schema, *a, **kw: schema.mutation_type,
        ),
        Field(
            "subscriptionType",
            __Type__,
            description=(
                "If this server supports subscription, the type that "
                "subscription operations will be rooted at."
            ),
            resolve=lambda schema, *a, **kw: schema.subscription_type,
        ),
        Field(
            "directives",
            NonNullType(ListType(NonNullType(__Directive__))),
            description="A list of all directives supported by this server.",
            resolve=lambda schema, *a, **kw: list(
                sorted(schema.directives.values(), key=lambda d: d.name)
            ),
        ),
    ],
)


__Directive__ = ObjectType(
    "__Directive",
    description=(
        "A Directive provides a way to describe alternate runtime execution "
        "and type validation behavior in a GraphQL document."
        "\n\nIn some cases, you need to provide options to alter GraphQL's "
        "execution behavior in ways field arguments will not suffice, such as "
        "conditionally including or skipping a field. Directives provide this "
        "by describing additional information to the executor."
    ),
    fields=lambda: [
        Field("name", NonNullType(String)),
        Field("description", String),
        Field(
            "locations",
            NonNullType(ListType(NonNullType(__DirectiveLocation__))),
        ),
        Field("args", NonNullType(ListType(NonNullType(__InputValue__)))),
    ],
)


__DirectiveLocation__ = EnumType(
    "__DirectiveLocation",
    [
        EnumValue(
            "QUERY", description="Location adjacent to a query operation."
        ),
        EnumValue(
            "MUTATION", description="Location adjacent to a mutation operation."
        ),
        EnumValue(
            "SUBSCRIPTION",
            description="Location adjacent to a subscription operation.",
        ),
        EnumValue("FIELD", description="Location adjacent to a field."),
        EnumValue(
            "FRAGMENT_DEFINITION",
            description="Location adjacent to a fragment definition.",
        ),
        EnumValue(
            "FRAGMENT_SPREAD",
            description="Location adjacent to a fragment spread.",
        ),
        EnumValue(
            "INLINE_FRAGMENT",
            description="Location adjacent to an inline fragment.",
        ),
        EnumValue(
            "SCHEMA", description="Location adjacent to a schema definition."
        ),
        EnumValue(
            "SCALAR", description="Location adjacent to a scalar definition."
        ),
        EnumValue(
            "OBJECT",
            description="Location adjacent to an object type definition.",
        ),
        EnumValue(
            "FIELD_DEFINITION",
            description="Location adjacent to a field definition.",
        ),
        EnumValue(
            "ARGUMENT_DEFINITION",
            description="Location adjacent to an argument definition.",
        ),
        EnumValue(
            "INTERFACE",
            description="Location adjacent to an interface definition.",
        ),
        EnumValue(
            "UNION", description="Location adjacent to a union definition."
        ),
        EnumValue(
            "ENUM", description="Location adjacent to an enum definition."
        ),
        EnumValue(
            "ENUM_VALUE",
            description="Location adjacent to an enum value definition.",
        ),
        EnumValue(
            "INPUT_OBJECT",
            description=(
                "Location adjacent to an input object type definition."
            ),
        ),
        EnumValue(
            "INPUT_FIELD_DEFINITION",
            description=(
                "Location adjacent to an input object field definition."
            ),
        ),
    ],
    description=(
        "A Directive can be adjacent to many parts of the GraphQL language, a "
        "__DirectiveLocation describes one such possible adjacencies."
    ),
)


def _resolve_type_kind(type_, *args, **kw):
    if isinstance(type_, ScalarType):
        return "SCALAR"
    elif isinstance(type_, ObjectType):
        return "OBJECT"
    elif isinstance(type_, InterfaceType):
        return "INTERFACE"
    elif isinstance(type_, UnionType):
        return "UNION"
    elif isinstance(type_, EnumType):
        return "ENUM"
    elif isinstance(type_, InputObjectType):
        return "INPUT_OBJECT"
    elif isinstance(type_, ListType):
        return "LIST"
    elif isinstance(type_, NonNullType):
        return "NON_NULL"
    raise TypeError("Unknown kind of type: %s" % type_)


__Type__ = ObjectType(
    "__Type",
    description=(
        "The fundamental unit of any GraphQL Schema is the type. There are "
        "many kinds of types in GraphQL as represented by the `__TypeKind` "
        "enum.\n\nDepending on the kind of a type, certain fields describe "
        "information about that type. Scalar types provide no information "
        "beyond a name and description, while Enum types provide their values. "
        "Object and Interface types provide the fields they describe. Abstract "
        "types, Union and Interface, provide the Object types possible "
        "at runtime. List and NonNull types compose other types."
    ),
    fields=lambda: [
        Field("kind", NonNullType(__TypeKind__), resolve=_resolve_type_kind),
        Field("name", String),
        Field("description", String),
        Field(
            "fields",
            ListType(NonNullType(__Field__)),
            args=[Argument("includeDeprecated", Boolean, default_value=False)],
            resolve=lambda type_, args, *a, **kw: (
                [
                    f
                    for f in type_.fields
                    if (not f.deprecated) or args.get("includeDeprecated")
                ]
                if isinstance(type_, (ObjectType, InterfaceType))
                else None
            ),
        ),
        Field(
            "interfaces",
            ListType(NonNullType(__Type__)),
            resolve=lambda type_, *a, **kw: (
                type_.interfaces if isinstance(type_, ObjectType) else None
            ),
        ),
        Field(
            "possibleTypes",
            ListType(NonNullType(__Type__)),
            resolve=lambda type_, a, c, info: (
                list(
                    sorted(
                        info.schema.get_possible_types(type_),
                        key=lambda t: t.name,
                    )
                )
                if is_abstract_type(type_)
                else None
            ),
        ),
        Field(
            "enumValues",
            ListType(NonNullType(__EnumValue__)),
            args=[Argument("includeDeprecated", Boolean, default_value=False)],
            resolve=lambda type_, args, *a, **kw: (
                [
                    ev
                    for ev in type_.values
                    if (not ev.deprecated) or args.get("includeDeprecated")
                ]
                if isinstance(type_, EnumType)
                else None
            ),
        ),
        Field(
            "inputFields",
            ListType(NonNullType(__InputValue__)),
            resolve=lambda type_, *a, **kw: (
                [f for f in type_.fields]
                if isinstance(type_, InputObjectType)
                else None
            ),
        ),
        Field(
            "ofType",
            __Type__,
            resolve=lambda type_, *a, **kw: (
                type_.type if isinstance(type_, WrappingType) else None
            ),
        ),
    ],
)


__EnumValue__ = ObjectType(
    "__EnumValue",
    description=(
        "One possible value for a given Enum. Enum values are unique values, "
        "not a placeholder for a string or numeric value. However an Enum "
        "value is returned in a JSON response as a string."
    ),
    fields=lambda: [
        Field("name", NonNullType(String)),
        Field("description", String),
        Field(
            "isDeprecated",
            NonNullType(Boolean),
            resolve=lambda ev, *_: ev.deprecated,
        ),
        Field(
            "deprecationReason",
            String,
            resolve=lambda ev, *_: ev.deprecation_reason,
        ),
    ],
)


def _format_default_value(input_value):
    if not input_value.has_default_value:
        return None
    dv = input_value.default_value
    if isinstance(dv, bool):
        return str(dv).lower()
    elif dv is None:
        return "null"
    elif isinstance(dv, six.string_types):
        return '"%s"' % dv
    return json.dumps(dv)


__InputValue__ = ObjectType(
    "__InputValue",
    description=(
        "Arguments provided to Fields or Directives and the input fields "
        "of an InputObject are represented as Input Values which describe "
        "their type and optionally a default value."
    ),
    fields=lambda: [
        Field("name", NonNullType(String)),
        Field("description", String),
        Field("type", NonNullType(__Type__)),
        Field(
            "defaultValue",
            String,
            description=(
                "A GraphQL-formatted string representing the "
                "default value for this input value."
            ),
            resolve=lambda iv, *a, **kw: _format_default_value(iv),
        ),
    ],
)


__Field__ = ObjectType(
    "__Field",
    description=(
        "Object and Interface types are described by a list of Fields, "
        "each of which has a name, potentially a list of arguments, and "
        "a return type."
    ),
    fields=lambda: [
        Field("name", NonNullType(String)),
        Field("description", String),
        Field(
            "args",
            NonNullType(ListType(NonNullType(__InputValue__))),
            resolve=lambda field, *args, **kwargs: (field.args or []),
        ),
        Field("type", NonNullType(__Type__)),
        Field(
            "isDeprecated",
            NonNullType(Boolean),
            resolve=lambda field, *_: field.deprecated,
        ),
        Field(
            "deprecationReason",
            String,
            resolve=lambda field, *_: field.deprecation_reason,
        ),
    ],
)


__TypeKind__ = EnumType(
    "__TypeKind",
    [
        EnumValue("SCALAR", description="Indicates this type is a scalar."),
        EnumValue(
            "OBJECT",
            description=(
                "Indicates this type is an object. `fields` and "
                "`interfaces` are valid fields."
            ),
        ),
        EnumValue(
            "INTERFACE",
            description=(
                "Indicates this type is an interface. `fields` and "
                "`possibleTypes` are valid fields."
            ),
        ),
        EnumValue(
            "UNION",
            description=(
                "Indicates this type is a union. `possibleTypes` is "
                "a valid field."
            ),
        ),
        EnumValue(
            "ENUM",
            description=(
                "Indicates this type is an enum. `enumValues` is a "
                "valid field."
            ),
        ),
        EnumValue(
            "INPUT_OBJECT",
            description=(
                "Indicates this type is an input object. `inputFields` "
                "is a valid field."
            ),
        ),
        EnumValue(
            "LIST",
            description=(
                "Indicates this type is a list. `ofType` is a valid " "field."
            ),
        ),
        EnumValue(
            "NON_NULL",
            description=(
                "Indicates this type is a non-null. `ofType` is a "
                "valid field."
            ),
        ),
    ],
    description="An enum describing what kind of type a given `__Type` is.",
)


INTROPSPECTION_TYPES = (
    __Schema__,
    __Directive__,
    __DirectiveLocation__,
    __Type__,
    __Field__,
    __InputValue__,
    __EnumValue__,
    __TypeKind__,
)

schema_field = Field(
    "__schema",
    NonNullType(__Schema__),
    description="Access the current type schema of this server.",
    args=[],
    resolve=lambda p, a, c, info: info.schema,
)

type_field = Field(
    "__type",
    __Type__,
    description="Request the type information of a single type.",
    args=[Argument("name", NonNullType(String))],
    resolve=lambda p, args, c, info: info.schema.get_type(args["name"], None),
)


type_name_field = Field(
    "__typename",
    NonNullType(String),
    description="The name of the current Object type at runtime.",
    args=[],
    resolve=lambda p, a, c, info: info.parent_type.name,
)


def is_introspection_type(type_):
    return type_ in INTROPSPECTION_TYPES



# -*- coding: utf-8 -*-
""" Validation rules defined in `this section
<http://facebook.github.io/graphql/June2018/#sec-Validation>`_ of the
specification. """

from ..._string_utils import infer_suggestions, quoted_options_list
from ..._utils import DefaultOrderedDict, OrderedDict, deduplicate
from ...exc import UnknownType
from ...lang import ast as _ast, print_ast
from ...lang.visitor import SkipNode
from ...schema import (
    InterfaceType,
    NonNullType,
    ObjectType,
    UnionType,
    is_composite_type,
    is_input_type,
    is_leaf_type,
    unwrap_type,
)
from ..visitors import ValidationVisitor, VariablesCollector
from .overlapping_fields_can_be_merged import (  # noqa: F401
    OverlappingFieldsCanBeMergedChecker
)
from .values_of_correct_type import ValuesOfCorrectTypeChecker  # noqa: F401


class ExecutableDefinitionsChecker(ValidationVisitor):
    """ A GraphQL document is only valid for execution if all definitions
    are either operation or fragment definitions.

    Unnecessary if parser was run with ``allow_type_system=False``.
    """

    def enter_document(self, node):
        for definition in node.definitions:
            if not isinstance(definition, _ast.ExecutableDefinition):
                self.add_error(
                    'Definition "%s" is not executable'
                    % (
                        "schema"
                        if isinstance(definition, _ast.SchemaDefinition)
                        else definition.name.value
                    ),
                    [definition],
                )
                raise SkipNode()


class UniqueOperationNameChecker(ValidationVisitor):
    """ A GraphQL document is only valid if all defined operations have
    unique names. """

    def __init__(self, schema, type_info):
        super(UniqueOperationNameChecker, self).__init__(schema, type_info)
        self.names = set()

    def enter_operation_definition(self, node):
        op_name = node.name.value if node.name else node.operation
        if op_name in self.names:
            self.add_error('Duplicate operation "%s".' % op_name, [node])
            raise SkipNode()
        self.names.add(op_name)


class LoneAnonymousOperationChecker(ValidationVisitor):
    """ A GraphQL document is only valid if when it contains an anonymous
    operation (the query short-hand) that it contains only that one
    operation definition. """

    def enter_document(self, node):
        operation_definitions = [
            d
            for d in node.definitions
            if isinstance(d, _ast.OperationDefinition)
        ]

        has_anonymous = any((d.name is None for d in operation_definitions))
        if has_anonymous and len(operation_definitions) > 1:
            self.add_error(
                "The anonymous operation must be the only defined operation.",
                [node],
            )
            raise SkipNode()


class SingleFieldSubscriptionsChecker(ValidationVisitor):
    """ A GraphQL subscription is valid only if it contains a single
    root field. """

    def enter_operation_definition(self, node):
        if node.operation == "subscription":
            if len(node.selection_set.selections) != 1:
                self.add_error(
                    'Subscription "%s" must select only one top level field.'
                    % (node.name.value if node.name else ""),
                    [node],
                )


class KnownTypeNamesChecker(ValidationVisitor):
    """ A GraphQL document is only valid if referenced types (specifically
    variable definitions and fragment conditions) are defined by the
    type schema. """

    def _skip(self, _):
        raise SkipNode()

    # Ignore type system defs
    enter_object_type_definition = _skip
    enter_interface_type_definition = _skip
    enter_union_type_definition = _skip
    enter_input_object_type_definition = _skip

    def _enter_type_literal(self, node):
        try:
            self.schema.get_type_from_literal(node)
        except UnknownType as err:
            self.add_error('Unknown type "%s"' % str(err), [node])

    enter_named_type = _enter_type_literal
    enter_list_type = _enter_type_literal
    enter_non_null_type = _enter_type_literal


class FragmentsOnCompositeTypesChecker(ValidationVisitor):
    """ Fragments use a type condition to determine if they apply, since
    fragments can only be spread into a composite type (object, interface, or
    union), the type condition must also be a composite type. """

    def enter_inline_fragment(self, node):
        if node.type_condition:
            type_ = self.schema.get_type_from_literal(node.type_condition)
            if not is_composite_type(type_):
                self.add_error(
                    "Fragment type condition cannot be on non-composite "
                    'type "%s"' % (type_.name),
                    [node],
                )
                raise SkipNode()

    def enter_fragment_definition(self, node):
        type_ = self.schema.get_type_from_literal(node.type_condition)
        if not is_composite_type(type_):
            self.add_error(
                'Fragment "%s" type condition cannot be on non-composite '
                'type "%s"' % (node.name.value, type_.name),
                [node],
            )
            raise SkipNode()


class VariablesAreInputTypesChecker(ValidationVisitor):
    """ A GraphQL operation is only valid if all the variables it defines are of
    input types (scalar, enum, or input object). """

    def enter_variable_definition(self, node):
        try:
            type_ = self.schema.get_type_from_literal(node.type)
        except UnknownType:
            type_ = None
        if not is_input_type(type_):
            self.add_error(
                'Variable "$%s" must be input type' % node.variable.name.value,
                [node],
            )


class ScalarLeafsChecker(ValidationVisitor):
    """ A GraphQL document is valid only if all leaf fields (fields without
    sub selections) are of scalar or enum types. """

    def enter_field(self, node):
        type_ = self.type_info.type

        if is_leaf_type(unwrap_type(type_)) and node.selection_set:
            self.add_error(
                'Field "%s" cannot have a selection as type "%s" has no '
                "fields" % (node.name.value, type_),
                [node],
            )

        if is_composite_type(unwrap_type(type_)) and not node.selection_set:
            self.add_error(
                'Field "%s" of type "%s" must have a subselection'
                % (node.name.value, type_),
                [node],
            )


class FieldsOnCorrectTypeChecker(ValidationVisitor):
    """ A GraphQL document is only valid if all fields selected are defined by
    the parent type, or are an allowed meta field such as __typename. """

    def enter_field(self, node):
        if self.type_info.parent_type is None:
            return

        field_def = self.type_info.field
        if field_def is None:

            if isinstance(
                self.type_info.parent_type, (ObjectType, InterfaceType)
            ):
                fieldnames = [f.name for f in self.type_info.parent_type.fields]
                suggestions = infer_suggestions(node.name.value, fieldnames)
                if suggestions:
                    self.add_error(
                        'Cannot query field "%s" on type "%r", '
                        "did you mean %s?"
                        % (
                            node.name.value,
                            self.type_info.parent_type,
                            quoted_options_list(suggestions),
                        ),
                        [node],
                    )
                else:
                    self.add_error(
                        'Cannot query field "%s" on type "%r"'
                        % (node.name.value, self.type_info.parent_type),
                        [node],
                    )

            elif isinstance(self.type_info.parent_type, UnionType):
                self.add_error(
                    'Cannot query field "%s" on type "%r", '
                    "did you mean to use an inline fragment on %s?"
                    % (
                        node.name.value,
                        self.type_info.parent_type,
                        quoted_options_list(
                            [t.name for t in self.type_info.parent_type.types]
                        ),
                    ),
                    [node],
                )

            else:
                self.add_error(
                    'Cannot query field "%s" on type "%r"'
                    % (node.name.value, self.type_info.parent_type),
                    [node],
                )


class UniqueFragmentNamesChecker(ValidationVisitor):
    """ A GraphQL document is only valid if all defined fragments have unique
    names. """

    def __init__(self, schema, type_info):
        super(UniqueFragmentNamesChecker, self).__init__(schema, type_info)
        self._names = set()

    def enter_fragment_definition(self, node):
        name = node.name.value
        if name in self._names:
            self.add_error(
                'There can only be one fragment named "%s"' % name, [node]
            )
        self._names.add(name)


class KnownFragmentNamesChecker(ValidationVisitor):
    """ A GraphQL document is only valid if all `...Fragment` fragment spreads
    refer to fragments defined in the same document. """

    def __init__(self, schema, type_info):
        super(KnownFragmentNamesChecker, self).__init__(schema, type_info)

    def enter_document(self, node):
        self._fragment_names = set(
            [
                definition.name.value
                for definition in node.definitions
                if type(definition) == _ast.FragmentDefinition
            ]
        )

    def enter_fragment_spread(self, node):
        name = node.name.value
        if name not in self._fragment_names:
            self.add_error('Unknown fragment "%s"' % name)


class NoUnusedFragmentsChecker(ValidationVisitor):
    """ A GraphQL document is only valid if all fragment definitions are spread
    within operations, or spread within other fragments spread within
    operations. """

    def __init__(self, schema, type_info):
        super(NoUnusedFragmentsChecker, self).__init__(schema, type_info)
        self._fragments = set()
        self._used_fragments = set()

    def enter_fragment_definition(self, node):
        self._fragments.add(node.name.value)

    def enter_fragment_spread(self, node):
        self._used_fragments.add(node.name.value)

    def leave_document(self, node):
        unused = self._fragments - self._used_fragments
        if unused:
            self.add_error(
                "Unused fragment(s) %s"
                % ", ".join(['"%s"' % n for n in sorted(unused)])
            )


class PossibleFragmentSpreadsChecker(ValidationVisitor):
    """ A fragment spread is only valid if the type condition could ever
    possibly be true: if there is a non-empty intersection of the possible
    parent types, and possible types which pass the type condition. """

    def __init__(self, schema, type_info):
        super(PossibleFragmentSpreadsChecker, self).__init__(schema, type_info)
        self._fragment_types = dict()

    def enter_document(self, node):
        self._fragment_types.update(
            {
                definition.name.value: self.schema.get_type_from_literal(
                    definition.type_condition
                )
                for definition in node.definitions
                if type(definition) == _ast.FragmentDefinition
            }
        )

    def enter_fragment_spread(self, node):
        name = node.name.value
        frag_type = self._fragment_types.get(name, None)
        parent_type = self.type_info.type

        if (
            is_composite_type(frag_type)
            and is_composite_type(parent_type)
            and not self.schema.types_overlap(frag_type, parent_type)
        ):
            self.add_error(
                'Fragment "%s" cannot be spread here as types "%s" and "%s"'
                " do not overlap." % (name, frag_type, parent_type),
                [node],
            )
            raise SkipNode()

    def enter_inline_fragment(self, node):
        type_ = self.type_info.type
        parent_type = self.type_info.parent_type

        if (
            is_composite_type(type_)
            and is_composite_type(parent_type)
            and not self.schema.types_overlap(type_, parent_type)
        ):
            self.add_error(
                'Inline fragment cannot be spread here as types "%s" and "%s"'
                " do not overlap." % (type_, parent_type),
                [node],
            )
            raise SkipNode()


class NoFragmentCyclesChecker(ValidationVisitor):
    """ A GraphQL Document is only valid if fragment definitions are not cyclic.
    """

    def __init__(self, schema, type_info):
        super(NoFragmentCyclesChecker, self).__init__(schema, type_info)
        self._spreads = OrderedDict()
        self._current = None

    def enter_fragment_definition(self, node):
        name = node.name.value
        self._current = name
        self._spreads[name] = []

    def leave_fragment_definition(self, node):
        self._current = None

    def enter_fragment_spread(self, node):
        name = node.name.value

        if self._current is not None:
            if self._current and name == self._current:
                self.add_error(
                    'Cannot spread fragment "%s" withing itself' % name, [node]
                )
                raise SkipNode()

            if name not in self._spreads[self._current]:
                self._spreads[self._current].append(name)

    def leave_document(self, node):
        def _search(outer, acc=None, path=None):
            acc, path = acc or dict(), path or []

            if outer not in self._spreads:
                return acc

            for inner in self._spreads[outer]:
                # The ref implementation will report multiple distinct cycles
                # for one fragment. This line and the fact that we keep one
                # path per fragment make it so that we only report one.
                if inner in acc:
                    break
                acc[inner] = path
                _search(inner, acc, path + [inner])

            return acc

        flat_spreads = [(outer, _search(outer)) for outer in self._spreads]
        cyclic = set()

        for outer, inner_spreads in flat_spreads:
            if outer in inner_spreads:
                cyclic.add(outer)
                path = inner_spreads[outer]

                # Make sure we do not report redundant cycles, i.e. if A > B
                # has been identified, A > B > A is redundant and will not be
                # reported.
                if path[-1] in cyclic:
                    continue

                self.add_error(
                    'Cannot spread fragment "%s" withing itself (via: %s)'
                    % (outer, " > ".join(path)),
                    [node],
                )


class UniqueVariableNamesChecker(ValidationVisitor):
    """ A GraphQL operation is only valid if all its variables are uniquely
    named. """

    def enter_operation_definition(self, node):
        self._variables = set()

    def leave_operation_definition(self, node):
        self._variables = None

    def enter_variable_definition(self, node):
        name = node.variable.name.value
        if name in self._variables:
            self.add_error('Duplicate variable "$%s"' % name, [node])
        self._variables.add(name)


class NoUndefinedVariablesChecker(VariablesCollector):
    """ A GraphQL operation is only valid if all variables encountered, both
    directly and via fragment spreads, are defined by that operation. """

    def leave_document(self, node):
        super(NoUndefinedVariablesChecker, self).leave_document(node)

        for op, fragments in self._op_fragments.items():
            defined = self._op_defined_variables[op]
            for fragment in deduplicate(fragments):
                fragment_variables = self._fragment_variables[fragment].items()
                for var, (node, _, _) in fragment_variables:
                    if var not in defined:
                        self.add_error(
                            'Variable "$%s" from fragment "%s" is not defined '
                            "on %s operation"
                            % (
                                var,
                                fragment,
                                '"%s"' % op if op != "" else "anonymous",
                            ),
                            [node],
                        )

        for op, variables in self._op_variables.items():
            defined = self._op_defined_variables[op]
            for var, (node, _, _) in variables.items():
                if var not in defined:
                    self.add_error(
                        'Variable "$%s" is not defined on %s operation'
                        % (var, '"%s"' % op if op != "" else "anonymous"),
                        [node],
                    )


class NoUnusedVariablesChecker(VariablesCollector):
    """ A GraphQL operation is only valid if all variables defined by an
    operation are used, either directly or within a spread fragment. """

    def leave_document(self, node):
        super(NoUnusedVariablesChecker, self).leave_document(node)

        used_variables = DefaultOrderedDict(set)

        for op, fragments in self._op_fragments.items():
            for fragment in deduplicate(fragments):
                for var in self._fragment_variables[fragment].keys():
                    used_variables[op].add(var)

        for op, variables in self._op_variables.items():
            for var in variables.keys():
                used_variables[op].add(var)

        for op, defined in self._op_defined_variables.items():
            used = used_variables[op]
            for var, node in defined.items():
                if var not in used:
                    self.add_error('Unused variable "$%s"' % var, [node])


class KnownDirectivesChecker(ValidationVisitor):
    """ A GraphQL document is only valid if all `@directives` are known by the
    schema and legally positioned. """

    def __init__(self, schema, type_info):
        super(KnownDirectivesChecker, self).__init__(schema, type_info)
        self._ancestors = []

    def _enter_ancestor(self, node):
        self._ancestors.append(node)

    def _leave_ancestor(self, node):
        self._ancestors.pop()

    enter_operation_definition = _enter_ancestor
    leave_operation_definition = _leave_ancestor
    enter_field = _enter_ancestor
    leave_field = _leave_ancestor
    enter_field = _enter_ancestor
    leave_field = _leave_ancestor
    enter_fragment_spread = _enter_ancestor
    leave_fragment_spread = _leave_ancestor
    enter_inline_fragment = _enter_ancestor
    leave_inline_fragment = _leave_ancestor
    enter_fragment_definition = _enter_ancestor
    leave_fragment_definition = _leave_ancestor
    enter_schema_definition = _enter_ancestor
    leave_schema_definition = _leave_ancestor
    enter_scalar_type_definition = _enter_ancestor
    leave_scalar_type_definition = _leave_ancestor
    enter_scalar_type_extension = _enter_ancestor
    leave_scalar_type_extension = _leave_ancestor
    enter_object_type_definition = _enter_ancestor
    leave_object_type_definition = _leave_ancestor
    enter_object_type_extension = _enter_ancestor
    leave_object_type_extension = _leave_ancestor
    enter_field_definition = _enter_ancestor
    leave_field_definition = _leave_ancestor
    enter_interface_type_definition = _enter_ancestor
    leave_interface_type_definition = _leave_ancestor
    enter_interface_type_extension = _enter_ancestor
    leave_interface_type_extension = _leave_ancestor
    enter_union_type_definition = _enter_ancestor
    leave_union_type_definition = _leave_ancestor
    enter_union_type_extension = _enter_ancestor
    leave_union_type_extension = _leave_ancestor
    enter_enum_type_definition = _enter_ancestor
    leave_enum_type_definition = _leave_ancestor
    enter_enum_type_extension = _enter_ancestor
    leave_enum_type_extension = _leave_ancestor
    enter_enum_value_definition = _enter_ancestor
    leave_enum_value_definition = _leave_ancestor
    enter_input_object_type_definition = _enter_ancestor
    leave_input_object_type_definition = _leave_ancestor
    enter_input_object_type_extension = _enter_ancestor
    leave_input_object_type_extension = _leave_ancestor
    enter_input_value_definition = _enter_ancestor
    leave_input_value_definition = _leave_ancestor

    def _current_location(self):
        ancestor = self._ancestors[-1]
        kind = type(ancestor)
        if kind is _ast.OperationDefinition:
            return {
                "query": "QUERY",
                "mutation": "MUTATION",
                "subscription": "SUBSCRIPTION",
            }.get(ancestor.operation, "QUERY")

        if kind is _ast.InputValueDefinition:
            parent = self._ancestors[-2]
            return (
                "INPUT_FIELD_DEFINITION"
                if type(parent) is _ast.InputObjectTypeDefinition
                else "ARGUMENT_DEFINITION"
            )

        return {
            _ast.Field: "FIELD",
            _ast.FragmentSpread: "FRAGMENT_SPREAD",
            _ast.InlineFragment: "INLINE_FRAGMENT",
            _ast.FragmentDefinition: "FRAGMENT_DEFINITION",
            _ast.SchemaDefinition: "SCHEMA",
            _ast.ScalarTypeDefinition: "SCALAR",
            _ast.ScalarTypeExtension: "SCALAR",
            _ast.ObjectTypeDefinition: "OBJECT",
            _ast.ObjectTypeExtension: "OBJECT",
            _ast.FieldDefinition: "FIELD_DEFINITION",
            _ast.InterfaceTypeDefinition: "INTERFACE",
            _ast.InterfaceTypeExtension: "INTERFACE",
            _ast.UnionTypeDefinition: "UNION",
            _ast.UnionTypeExtension: "UNION",
            _ast.EnumTypeDefinition: "ENUM",
            _ast.EnumTypeExtension: "ENUM",
            _ast.EnumValueDefinition: "ENUM_VALUE",
            _ast.InputObjectTypeDefinition: "INPUT_OBJECT",
            _ast.InputObjectTypeExtension: "INPUT_OBJECT",
        }[kind]

    def enter_directive(self, node):
        name = node.name.value
        schema_directive = self.schema.directives.get(name)
        if schema_directive is None:
            self.add_error('Unknown directive "@%s"' % name, [node])
        else:
            location = self._current_location()
            if location not in schema_directive.locations:
                self.add_error(
                    'Directive "@%s" may not be used on %s' % (name, location),
                    [node],
                )


class UniqueDirectivesPerLocationChecker(ValidationVisitor):
    """ A GraphQL document is only valid if all directives at a given location
    are uniquely named. """

    def _validate_unique_directive_names(self, node):
        seen = set()
        for directive in node.directives:
            name = directive.name.value
            if name in seen:
                self.add_error('Duplicate directive "@%s"' % name, [directive])
            seen.add(name)

    enter_operation_definition = _validate_unique_directive_names
    enter_field = _validate_unique_directive_names
    enter_field = _validate_unique_directive_names
    enter_fragment_spread = _validate_unique_directive_names
    enter_inline_fragment = _validate_unique_directive_names
    enter_fragment_definition = _validate_unique_directive_names


class KnownArgumentNamesChecker(ValidationVisitor):
    """ A GraphQL field / directive is only valid if all supplied arguments
    are defined by that field / directive. """

    def enter_field(self, node):
        field_def = self.type_info.field
        if field_def is not None:
            known = set((a.name for a in field_def.args))
            for arg in node.arguments:
                name = arg.name.value
                if name not in known:
                    suggestions = list(infer_suggestions(name, known))
                    if not suggestions:
                        self.add_error(
                            'Unknown argument "%s" on field "%s" of type "%s"'
                            % (
                                name,
                                field_def.name,
                                self.type_info.parent_type,
                            ),
                            [arg],
                        )
                    else:
                        self.add_error(
                            'Unknown argument "%s" on field "%s" of type "%s", '
                            "did you mean %s?"
                            % (
                                name,
                                field_def.name,
                                self.type_info.parent_type,
                                quoted_options_list(suggestions),
                            ),
                            [arg],
                        )

    def enter_directive(self, node):
        directive_def = self.type_info.directive
        if directive_def is not None:
            known = set((a.name for a in directive_def.args))
            for arg in node.arguments:
                name = arg.name.value
                if name not in known:
                    suggestions = infer_suggestions(name, known)
                    if not suggestions:
                        self.add_error(
                            'Unknown argument "%s" on directive "@%s"'
                            % (name, directive_def.name),
                            [arg],
                        )
                    else:
                        self.add_error(
                            'Unknown argument "%s" on directive "@%s", did you mean %s?'
                            % (
                                name,
                                directive_def.name,
                                quoted_options_list(suggestions),
                            ),
                            [arg],
                        )


class UniqueArgumentNamesChecker(ValidationVisitor):
    """ A GraphQL field or directive is only valid if all supplied arguments
    are uniquely named. """

    def _check_duplicate_args(self, node):
        argnames = set()
        for arg in node.arguments:
            name = arg.name.value
            if name in argnames:
                self.add_error('Duplicate argument "%s"' % name, [arg])
            argnames.add(name)

    enter_field = _check_duplicate_args
    enter_directive = _check_duplicate_args


class ProvidedRequiredArgumentsChecker(ValidationVisitor):
    """ A field or directive is only valid if all required (non-null without a
    default value) ) field arguments have been provided. """

    def _missing_args(self, arg_defs, node):
        node_args = set((arg.name.value for arg in node.arguments))
        for arg in arg_defs:
            if arg.required and (arg.name not in node_args):
                yield arg

    # Validate on leave to allow for deeper errors to appear first.
    def leave_field(self, node):
        field_def = self.type_info.field
        if field_def:
            for arg in self._missing_args(field_def.args, node):
                self.add_error(
                    'Field "%s" argument "%s" of type %s is required but '
                    "not provided" % (field_def.name, arg.name, arg.type),
                    [node],
                )

    # Validate on leave to allow for deeper errors to appear first.
    def leave_directive(self, node):
        directive_def = self.type_info.directive
        if directive_def:
            for arg in self._missing_args(directive_def.args, node):
                self.add_error(
                    'Directive "@%s" argument "%s" of type %s is required but '
                    "not provided" % (directive_def.name, arg.name, arg.type),
                    [node],
                )


class VariablesInAllowedPositionChecker(VariablesCollector):
    """ Variables passed to field arguments conform to type """

    def iter_op_variables(self, op):
        for usage in self._op_variables[op].items():
            yield usage
        for fragment in self._op_fragments[op]:
            frament_vars = self._fragment_variables[fragment].items()
            for usage in frament_vars:
                yield usage

    def leave_document(self, node):
        super(VariablesInAllowedPositionChecker, self).leave_document(node)

        for op, vardefs in self._op_defined_variables.items():

            for (varname, usage) in self.iter_op_variables(op):
                varnode, input_type, input_value_def = usage
                vardef = vardefs.get(varname)
                if vardef and input_type:

                    try:
                        var_type = self.schema.get_type_from_literal(
                            vardef.type
                        )
                    except UnknownType:
                        continue

                    var_default = vardef.default_value

                    if isinstance(input_type, NonNullType) and not isinstance(
                        var_type, NonNullType
                    ):
                        non_null_var_default = (
                            var_default is not None
                            and type(var_default) != _ast.NullValue
                        )
                        location_default = (
                            input_value_def is not None
                            and input_value_def.has_default_value
                        )
                        if (
                            not non_null_var_default and not location_default
                        ) or not self.schema.is_subtype(
                            var_type, input_type.type
                        ):
                            self.add_error(
                                'Variable "$%s" of type %s used in position '
                                "expecting type %s"
                                % (varname, print_ast(vardef.type), input_type),
                                [varnode],
                            )
                    else:
                        if not self.schema.is_subtype(var_type, input_type):
                            self.add_error(
                                'Variable "$%s" of type %s used in position '
                                "expecting type %s"
                                % (varname, print_ast(vardef.type), input_type),
                                [varnode],
                            )


class UniqueInputFieldNamesChecker(ValidationVisitor):
    """ A GraphQL input object value is only valid if all supplied fields are
    uniquely named.
    """

    def __init__(self, schema, type_info):
        super(UniqueInputFieldNamesChecker, self).__init__(schema, type_info)
        self._stack = []

    def enter_object_value(self, node):
        self._stack.append((node, set()))

    def leave_object_value(self, node):
        self._stack.pop()

    def enter_object_field(self, node):
        fieldname = node.name.value
        _, names = self._stack[-1]
        if fieldname in names:
            self.add_error(
                "There can be only one input field named %s." % fieldname,
                [node],
            )
        names.add(fieldname)


__all__ = (
    "ExecutableDefinitionsChecker",
    "UniqueOperationNameChecker",
    "LoneAnonymousOperationChecker",
    "SingleFieldSubscriptionsChecker",
    "KnownTypeNamesChecker",
    "FragmentsOnCompositeTypesChecker",
    "VariablesAreInputTypesChecker",
    "ScalarLeafsChecker",
    "FieldsOnCorrectTypeChecker",
    "UniqueFragmentNamesChecker",
    "KnownFragmentNamesChecker",
    "NoUnusedFragmentsChecker",
    "PossibleFragmentSpreadsChecker",
    "NoFragmentCyclesChecker",
    "UniqueVariableNamesChecker",
    "NoUndefinedVariablesChecker",
    "NoUnusedVariablesChecker",
    "KnownDirectivesChecker",
    "UniqueDirectivesPerLocationChecker",
    "KnownArgumentNamesChecker",
    "UniqueArgumentNamesChecker",
    "ValuesOfCorrectTypeChecker",
    "ProvidedRequiredArgumentsChecker",
    "VariablesInAllowedPositionChecker",
    "OverlappingFieldsCanBeMergedChecker",
    "UniqueInputFieldNamesChecker",
)



# -*- coding: utf-8 -*-
""" Test schema validation """

import pytest

from py_gql.exc import SchemaError
from py_gql.schema import (
    Argument,
    EnumType,
    EnumValue,
    Field,
    InputField,
    InputObjectType,
    Int,
    InterfaceType,
    ListType,
    NonNullType,
    ObjectType,
    ScalarType,
    Schema,
    String,
    UnionType,
)
from py_gql.schema._validation import validate_schema

SomeScalar = ScalarType(
    "SomeScalar",
    serialize=lambda a: None,
    parse=lambda a: None,
    parse_literal=lambda a, **k: None,
)

SomeObject = ObjectType("SomeObject", [Field("f", String)])

IncompleteObject = ObjectType("IncompleteObject", [])

SomeUnion = UnionType("SomeUnion", [SomeObject])

SomeInterface = InterfaceType("SomeInterface", [Field("f", String)])

SomeEnum = EnumType("SomeEnum", [EnumValue("ONLY")])

SomeInputObject = InputObjectType(
    "SomeInputObject", [InputField("val", String, default_value="hello")]
)


def _type_modifiers(t):
    return [t, ListType(t), NonNullType(t), NonNullType(ListType(t))]


def _with_modifiers(types):
    out = []
    for t in types:
        out.extend(_type_modifiers(t))
    return out


output_types = _with_modifiers(
    [String, SomeScalar, SomeEnum, SomeObject, SomeUnion, SomeInterface]
)

not_output_types = _with_modifiers([SomeInputObject])

input_types = _with_modifiers([String, SomeScalar, SomeEnum, SomeInputObject])

not_input_types = _with_modifiers([SomeObject, SomeUnion, SomeInterface])


def _single_type_schema(type_, fieldname="f"):
    return Schema(ObjectType("Query", [Field(fieldname, type_)]), types=[type_])


def test_query_type_is_an_object_type():
    schema = _single_type_schema(String, "test")
    assert validate_schema(schema)


def test_query_and_mutation_types_are_object_types():
    schema = Schema(
        ObjectType("Query", [Field("test", String)]),
        mutation_type=ObjectType("Mutation", [Field("test", String)]),
    )
    assert validate_schema(schema)


def test_query_and_subscription_types_are_object_types():
    schema = Schema(
        ObjectType("Query", [Field("test", String)]),
        subscription_type=ObjectType("Subscription", [Field("test", String)]),
    )
    assert validate_schema(schema)


def test_reject_non_object_query_type():
    schema = Schema(String)

    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert str(exc_info.value) == 'Query must be ObjectType but got "String"'


def test_reject_non_object_mutation_type():
    schema = Schema(
        ObjectType("Query", [Field("test", String)]), mutation_type=String
    )

    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert str(exc_info.value) == 'Mutation must be ObjectType but got "String"'


def test_reject_non_object_subscription_type():
    schema = Schema(
        ObjectType("Query", [Field("test", String)]), subscription_type=String
    )

    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Subscription must be ObjectType but got "String"'
    )


def test_reject_incorrectly_typed_directives():
    with pytest.raises(SchemaError) as exc_info:
        Schema(SomeObject, directives=["somedirective"])

    assert "somedirective" in str(exc_info.value)
    assert "str" in str(exc_info.value)


def test_accept_object_type_with_fields_object():
    schema = _single_type_schema(SomeObject)
    assert validate_schema(schema)


def test_reject_object_without_fields():
    schema = _single_type_schema(IncompleteObject)
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Type "IncompleteObject" must define at least one field'
    )


def test_reject_object_with_incorrectly_named_fields():
    schema = Schema(
        ObjectType("Query", [Field("bad-name-with-dashes", String)])
    )

    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value) == 'Invalid name "bad-name-with-dashes", '
        "must match /^[_a-zA-Z][_a-zA-Z0-9]*$/"
    )


def test_reject_incorrectly_named_type():
    schema = _single_type_schema(
        ObjectType("bad-name-with-dashes", [Field("field", String)])
    )

    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value) == 'Invalid type name "bad-name-with-dashes", '
        "must match /^[_a-zA-Z][_a-zA-Z0-9]*$/"
    )


def test_accept_field_args_with_correct_names():
    schema = _single_type_schema(
        ObjectType(
            "SomeObject",
            [Field("field", String, [Argument("goodArg", String)])],
        )
    )
    assert validate_schema(schema)


def test_reject_field_args_with_incorrect_names():
    schema = _single_type_schema(
        ObjectType(
            "SomeObject",
            [Field("field", String, [Argument("bad-arg", String)])],
        )
    )

    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Invalid name "bad-arg", must match /^[_a-zA-Z][_a-zA-Z0-9]*$/'
    )


def test_accept_union_type_with_valid_members():
    TypeA = ObjectType("TypeA", [Field("f", String)])
    TypeB = ObjectType("TypeB", [Field("f", String)])
    GoodUnion = UnionType("GoodUnion", [TypeA, TypeB])
    schema = _single_type_schema(GoodUnion)
    assert schema.validate()


def test_reject_union_type_with_no_member():
    EmptyUnion = UnionType("EmptyUnion", [])
    schema = _single_type_schema(EmptyUnion)

    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'UnionType "EmptyUnion" must at least define one member'
    )


def test_reject_union_type_with_duplicate_members():
    TypeA = ObjectType("TypeA", [Field("f", String)])
    BadUnion = UnionType("BadUnion", [TypeA, TypeA])
    schema = _single_type_schema(BadUnion)
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'UnionType "BadUnion" can only include type "TypeA" once'
    )


def test_reject_union_type_with_non_object_members():
    TypeA = ObjectType("TypeA", [Field("f", String)])
    BadUnion = UnionType("BadUnion", [TypeA, SomeScalar])
    schema = _single_type_schema(BadUnion)
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'UnionType "BadUnion" expects object types but got "SomeScalar"'
    )


def test_accept_input_type():
    schema = _single_type_schema(
        ObjectType(
            "Object",
            [Field("field", String, [Argument("arg", SomeInputObject)])],
        )
    )
    assert schema.validate()


def test_reject_input_type_with_no_fields():
    EmptyInput = InputObjectType("EmptyInput", [])
    schema = _single_type_schema(
        ObjectType(
            "Object", [Field("field", String, [Argument("arg", EmptyInput)])]
        )
    )

    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Type "EmptyInput" must define at least one field'
    )


def test_reject_input_type_with_incorectly_typed_fields():
    BadInput = InputObjectType("BadInput", [InputField("f", SomeObject)])
    schema = _single_type_schema(
        ObjectType(
            "Object", [Field("field", String, [Argument("arg", BadInput)])]
        )
    )

    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Expected input type for field "f" on "BadInput" '
        'but got "SomeObject"'
    )


def test_reject_enum_type_with_no_values():
    schema = _single_type_schema(EnumType("EmptyEnum", []))
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'EnumType "EmptyEnum" must at least define one value'
    )


def test_reject_enum_value_with_incorrect_name():
    schema = _single_type_schema(EnumType("BadEnum", ["#value"]))
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Invalid name "#value", must match /^[_a-zA-Z][_a-zA-Z0-9]*$/'
    )


@pytest.mark.parametrize("type_", output_types, ids=lambda t: "type = %s" % t)
def test_accept_output_type_as_object_fields(type_):
    schema = _single_type_schema(type_)
    assert schema.validate()


@pytest.mark.parametrize(
    "type_", not_output_types + [None], ids=lambda t: "type = %s" % t
)
def test_reject_non_output_type_as_object_fields(type_):
    schema = _single_type_schema(type_)
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Expected output type for field "f" on "Query" but got "%s"' % type_
    )


def test_reject_object_implementing_same_interface_twice():
    schema = _single_type_schema(
        ObjectType(
            "SomeObject",
            [Field("f", String)],
            interfaces=[SomeInterface, SomeInterface],
        )
    )
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value) == 'Type "SomeObject" mut only implement interface '
        '"SomeInterface" once'
    )


@pytest.mark.parametrize("type_", output_types, ids=lambda t: "type = %s" % t)
def test_accept_interface_fields_with_output_type(type_):
    iface = InterfaceType("GoodInterface", [Field("f", type_)])
    schema = _single_type_schema(iface)
    assert schema.validate()


@pytest.mark.parametrize(
    "type_", not_output_types + [None], ids=lambda t: "type = %s" % t
)
def test_reject_interface_fields_with_non_output_type(type_):
    iface = InterfaceType("BadInterface", [Field("f", type_)])
    schema = _single_type_schema(iface)
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Expected output type for field "f" on "BadInterface" '
        'but got "%s"' % type_
    )


def test_reject_interface_with_no_field():
    iface = InterfaceType("BadInterface", [])
    schema = _single_type_schema(iface)
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Type "BadInterface" must define at least one field'
    )


@pytest.mark.parametrize("type_", input_types, ids=lambda t: "type = %s" % t)
def test_accept_argument_with_input_type(type_):
    schema = _single_type_schema(
        ObjectType(
            "GoodObject", [Field("f", String, [Argument("goodArg", type_)])]
        )
    )
    assert schema.validate()


@pytest.mark.parametrize(
    "type_", not_input_types + [None], ids=lambda t: "type = %s" % t
)
def test_reject_argument_with_non_input_type(type_):
    schema = _single_type_schema(
        ObjectType(
            "BadObject", [Field("f", String, [Argument("badArg", type_)])]
        )
    )
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Expected input type for argument "badArg" on "BadObject.f" '
        'but got "%s"' % type_
    )


@pytest.mark.parametrize("type_", input_types, ids=lambda t: "type = %s" % t)
def test_accept_input_object_with_input_type(type_):
    schema = _single_type_schema(
        ObjectType(
            "GoodObject",
            [
                Field(
                    "f",
                    String,
                    [
                        Argument(
                            "goodArg",
                            InputObjectType(
                                "GoodInput", [InputField("f", type_)]
                            ),
                        )
                    ],
                )
            ],
        )
    )
    assert schema.validate()


@pytest.mark.parametrize(
    "type_", not_input_types + [None], ids=lambda t: "type = %s" % t
)
def test_reject_input_object_with_non_input_type(type_):
    schema = _single_type_schema(
        ObjectType(
            "BadObject",
            [
                Field(
                    "f",
                    String,
                    [
                        Argument(
                            "badArg",
                            InputObjectType(
                                "BadInput", [InputField("f", type_)]
                            ),
                        )
                    ],
                )
            ],
        )
    )
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Expected input type for field "f" on "BadInput" '
        'but got "%s"' % type_
    )


def test_reject_input_object_with_no_field():
    schema = _single_type_schema(
        ObjectType(
            "BadObject",
            [
                Field(
                    "f",
                    String,
                    [Argument("badArg", InputObjectType("BadInput", []))],
                )
            ],
        )
    )
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value) == 'Type "BadInput" must define at least one field'
    )


def test_accept_object_which_implements_interface():
    schema = _single_type_schema(
        ObjectType(
            "SomeObject", [Field("f", String)], interfaces=[SomeInterface]
        )
    )
    assert schema.validate()


def test_accept_object_which_implements_interface_along_with_more_fields():
    schema = _single_type_schema(
        ObjectType(
            "SomeObject",
            [Field("f", String), Field("f2", String)],
            interfaces=[SomeInterface],
        )
    )
    assert schema.validate()


def test_accept_object_which_implements_interface_along_with_nullable_args():
    schema = _single_type_schema(
        ObjectType(
            "SomeObject",
            [Field("f", String, [Argument("arg", String)])],
            interfaces=[SomeInterface],
        )
    )
    assert schema.validate()


def test_reject_object_missing_interface_field():
    schema = _single_type_schema(
        ObjectType(
            "SomeObject", [Field("_f", String)], interfaces=[SomeInterface]
        )
    )
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Interface field "SomeInterface.f" is not implemented '
        'by type "SomeObject"'
    )


def test_reject_object_with_incorrectly_typed_interface_field():
    schema = _single_type_schema(
        ObjectType("SomeObject", [Field("f", Int)], interfaces=[SomeInterface])
    )
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Interface field "SomeInterface.f" expects type "String" '
        'but "SomeObject.f" is type "Int"'
    )


def test_accept_object_fields_with_interface_subtype_of_interface_field():
    iface = InterfaceType("IFace", [Field("f", lambda: iface)])
    obj = ObjectType("Obj", [Field("f", lambda: obj)], interfaces=[iface])
    schema = _single_type_schema(obj)
    assert schema.validate()


def test_accept_object_fields_with_union_subtype_of_interface_field():
    union = UnionType("union", [SomeObject])
    iface = InterfaceType("IFace", [Field("f", union)])
    obj = ObjectType("Obj", [Field("f", SomeObject)], interfaces=[iface])
    schema = _single_type_schema(obj)
    assert schema.validate()


def test_reject_object_fields_with_missing_interface_argument():
    iface = InterfaceType(
        "IFace", [Field("f", String, [Argument("arg", String)])]
    )
    obj = ObjectType("Obj", [Field("f", String)], interfaces=[iface])
    schema = _single_type_schema(obj)
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value) == 'Interface field argument "IFace.f.arg" is not '
        'provided by "Obj.f"'
    )


def test_reject_object_fields_with_incorrectly_typed_interface_argument():
    iface = InterfaceType(
        "IFace", [Field("f", String, [Argument("arg", String)])]
    )
    obj = ObjectType(
        "Obj", [Field("f", String, [Argument("arg", Int)])], interfaces=[iface]
    )
    schema = _single_type_schema(obj)
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value) == 'Interface field argument "IFace.f.arg" expects '
        'type "String" but "Obj.f.arg" is type "Int"'
    )


def test_reject_object_which_implements_interface_along_with_required_args():
    iface = InterfaceType("IFace", [Field("f", String)])
    schema = _single_type_schema(
        ObjectType(
            "SomeObject",
            [Field("f", String, [Argument("arg", NonNullType(String))])],
            interfaces=[iface],
        )
    )
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Object field argument "SomeObject.f.arg" is of required '
        'type "String!" but is not provided by interface field "IFace.f"'
    )


def test_accept_object_with_list_interface_list_field():
    iface = InterfaceType("IFace", [Field("f", ListType(String))])
    schema = _single_type_schema(
        ObjectType(
            "SomeObject", [Field("f", ListType(String))], interfaces=[iface]
        )
    )
    assert schema.validate()


def test_accept_object_with_non_list_interface_list_field():
    iface = InterfaceType("IFace", [Field("f", ListType(String))])
    schema = _single_type_schema(
        ObjectType("SomeObject", [Field("f", String)], interfaces=[iface])
    )
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Interface field "IFace.f" expects type "[String]" but '
        '"SomeObject.f" is type "String"'
    )


def test_accept_object_with_list_interface_non_list_field():
    iface = InterfaceType("IFace", [Field("f", String)])
    schema = _single_type_schema(
        ObjectType(
            "SomeObject", [Field("f", ListType(String))], interfaces=[iface]
        )
    )
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Interface field "IFace.f" expects type "String" but '
        '"SomeObject.f" is type "[String]"'
    )


def test_accept_object_with_non_null_interface_null_field():
    iface = InterfaceType("IFace", [Field("f", String)])
    schema = _single_type_schema(
        ObjectType(
            "SomeObject", [Field("f", NonNullType(String))], interfaces=[iface]
        )
    )
    assert schema.validate()


def test_reject_object_with_null_interface_non_null_field():
    iface = InterfaceType("IFace", [Field("f", NonNullType(String))])
    schema = _single_type_schema(
        ObjectType("SomeObject", [Field("f", String)], interfaces=[iface])
    )
    with pytest.raises(SchemaError) as exc_info:
        validate_schema(schema)

    assert (
        str(exc_info.value)
        == 'Interface field "IFace.f" expects type "String!" but '
        '"SomeObject.f" is type "String"'
    )


def test_starwars_schema_is_valid(starwars_schema):
    assert validate_schema(starwars_schema)



# -*- coding: utf-8 -*-
""" Some generic laguage level utilities for internal use. """

import collections
import sys

import six


def lazy(maybe_callable):
    """ Calls a value if callable else returns it.

    >>> lazy(42)
    42

    >>> lazy(lambda: 42)
    42
    """
    if callable(maybe_callable):
        return maybe_callable()
    return maybe_callable


_MISSING = object()


class cached_property(property):
    """ Decorator that converts a method into a lazy property.
    The method behaves like a regular ``@property`` except that it
    is called only once.

    **Warning** This requires the class to have a ``__dict__`` attribute,
    so no class using __slots__.
    """

    # pylint: disable = super-init-not-called
    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = value

    # pylint: disable = redefined-builtin
    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _MISSING)
        if value is _MISSING:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


def _ident(x):
    return x


def deduplicate(iterable, key=_ident):
    """ Deduplicate an iterable.

    Args:
        iterable (Iterator[any]): source iterator
        key (Callable): Identity function

    Yields:
        any: next deduplicated entry in order of original appearance

    >>> list(deduplicate([1, 2, 1, 3, 3]))
    [1, 2, 3]

    >>> list(deduplicate([1, 2, 1, 3, 3], key=lambda x: x % 2))
    [1, 2]
    """
    seen = set()

    for entry in iterable:
        key_ = key(entry)
        if key_ in seen:
            continue
        seen.add(key_)
        yield entry


def maybe_first(iterable, default=None):
    """ Return the first item in an iterable or None.

    >>> maybe_first([1, 2, 3])
    1

    >>> maybe_first(()) is None
    True

    >>> maybe_first((), 1)
    1
    """
    try:
        return next(iter(iterable))
    except StopIteration:
        return default


def find_one(iterable, predicate, default=None):
    """ Extract first item matching a predicate function in an iterable.
    Returns ``None`` if no entry is found.

    >>> find_one([1, 2, 3, 4], lambda x: x == 2)
    2

    >>> find_one([1, 2, 3, 4], lambda x: x == 5) is None
    True
    """
    return maybe_first(
        (entry for entry in iterable if predicate(entry)), default=default
    )


if sys.version >= "3.7":  # flake8: noqa
    # Take advantage that dicts are guaranteed ordered from 3.7 onward
    OrderedDict = dict
    DefaultOrderedDict = collections.defaultdict
else:
    OrderedDict = collections.OrderedDict

    # Source: http://stackoverflow.com/a/6190500/562769
    class DefaultOrderedDict(OrderedDict):
        def __init__(self, default_factory=None, *a, **kw):
            if default_factory is not None and not callable(default_factory):
                raise TypeError("first argument must be callable")
            OrderedDict.__init__(self, *a, **kw)
            self.default_factory = default_factory

        def __getitem__(self, key):
            try:
                return OrderedDict.__getitem__(self, key)
            except KeyError:
                return self.__missing__(key)

        def __missing__(self, key):
            if self.default_factory is None:
                raise KeyError(key)
            self[key] = value = self.default_factory()
            return value

        def __reduce__(self):
            if self.default_factory is None:
                args = tuple()
            else:
                args = (self.default_factory,)
            return type(self), args, None, None, self.items()

        def copy(self):
            return self.__copy__()

        def __copy__(self):
            return type(self)(self.default_factory, self)

        def __deepcopy__(self, memo):
            import copy

            return type(self)(self.default_factory, copy.deepcopy(self.items()))

        def __repr__(self):
            return "DefaultOrderedDict(%s, %s)" % (
                self.default_factory,
                OrderedDict.__repr__(self),
            )


def flatten(lst):
    """ Recursive flatten (list, tuple) of potentially (lists, tuples)
    `itertools.chain` could be used instead but would flatten all iterables
    including strings which is not ideal.

    >>> list(flatten([1, 2, (1, 2, [3])]))
    [1, 2, 1, 2, 3]
    """
    for entry in lst:
        if type(entry) in (list, tuple):
            for subentry in flatten(entry):
                yield subentry
        else:
            yield entry


def is_iterable(value, strings=True):
    """ Check if a value is iterable.

    This does no type comparisons.
    Note that by default strings are iterables too!

    Args:
        value (any): Maybe iterable
        strings (bool): Include strings as iterable, defaults to ``True``

    Returns:
        bool: Whether ``value`` is iterable

    >>> is_iterable([])
    True

    >>> is_iterable("Hello World!")
    True

    >>> is_iterable("Hello World!", False)
    False

    >>> is_iterable(False)
    False
    """
    try:
        iter(value)
    except TypeError:
        return False
    else:
        return strings or not isinstance(value, six.string_types)


def nested_key(obj, *keys, **kwargs):
    """ Safely extract nested key from dicts and lists.

    >>> source = {'foo': {'bar': [{}, {}, {'baz': 42}]}}

    >>> nested_key(source, 'foo', 'bar', 2, 'baz')
    42

    >>> nested_key(source, 'foo', 'bar', 10, 'baz')
    Traceback (most recent call last):
        ...
    IndexError: list index out of range

    >>> nested_key(source, 'foo', 'bar', 10, 'baz', default=None) is None
    True

    >>> nested_key(source, 'foo', 'baz', default=None) is None
    True

    """
    default = kwargs.get("default", _MISSING)
    for key in keys:
        try:
            obj = obj[key]
        except (KeyError, IndexError):
            if default is _MISSING:
                raise
            return default
    return obj
